import fsspec
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
import dropbox


# TODO - still a lot to do here, not working just yet

# TODO - add support for other storage types, see: fsspec.available_protocols()

# TODO - check if its advantageous to use FUSE with fsspec: https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.fuse.run

class FileSystemClient(fsspec.AbstractFileSystem):
    # list of all available methods: https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.spec.AbstractFileSystem

    def __init__(self, fs_type, fs_kwargs):

        if fs_type == "local":
            super().__init__("file")
            self.prefix = ""

        elif fs_type == "s3":
            # ref: https://s3fs.readthedocs.io/en/latest/api.html#s3fs.core.S3FileSystem
            super().__init__(
                "s3", 
                anon=False,
                key=fs_kwargs.get("aws_access_key_id", ""),  # access key ID
                secret=fs_kwargs.get("aws_secret_access_key", ""),  # secret access key
                client_kwargs={
                    'endpoint_url': 'https://s3.amazonaws.com', 
                    'region_name': fs_kwargs.get("aws_region_name", "us-east-1")
                }
            )
            self.prefix = ""

        elif fs_type == "gdrive":
            # TODO - Create credentials object and authenticate with Google Drive API
            creds = Credentials.from_authorized_user()
            service = build("drive", "v3", credentials=creds)

            # Mount file system from Google Drive
            super().__init__("gcs", service=service, token="cache")
            self.prefix = "gdrive://"

        elif fs_type == "dropbox":
            # TODO - Create Dropbox client object
            client = dropbox.Dropbox(fs_kwargs)

            # Mount file system from Dropbox
            super().__init__("dropbox", client=client)
            self.prefix = ""

        else:
            raise ValueError(f"Unsupported file system type: {fs_type}")

    def read_file(self, path):
        with self.open(self.prefix + path) as f:
            return f.read()


from starflow.base_piece import BasePiece


def create_piece(
    metadata: dict, 
    piece_function: callable, 
    generate_report: callable = None
):
    class NewPiece(BasePiece): pass

    # Operator name as used by Airflow
    NewPiece.__name__ = metadata.get("name", "BasePiece")
    
    # Operator custom function
    NewPiece.piece_function = piece_function

    # Operator custom Generate Report function
    if generate_report:
        NewPiece.generate_report = generate_report
    
    return NewPiece


class InvalidPieceOutputError(Exception):
    """
    Raised when the Operator's output data is not an instance of the Operator's OutputModel
    """

    def __init__(self, piece_name: str):
        message = f"The output data for {piece_name} is not an instance of its OutputModel"
        super().__init__(message)


class MissingPieceRepositoryFileError(Exception):
    """
    Raised when there's a missing required file from Operators repository
    """

    def __init__(self, missing_file: str, pieces_repository: str):
        message = f"The file {missing_file} is not present in {pieces_repository} pieces repository"
        super().__init__(message)


class MissingEnvVarError(Exception):
    """
    Raised when there's a required ENV var missing on running environment
    """

    def __init__(self, missing_vars: list):
        message = "The required secrets are not present in running environment: " + '\n' + '\n'.join(missing_vars)
        super().__init__(message)


class NoMatchingDependencyForPieceError(Exception):
    """
    Raised when there's no matching dependency for the given Operator in the dependencies_map.json file
    """

    def __init__(self, piece_name: str, repo_name: str):
        message = f"There's no matching dependency group for {piece_name} from repository {repo_name}. Please make sure to run 'starflow organize' in the target repository."
        super().__init__(message)


import subprocess
from typing import List, Union, Tuple
import ast
import os
from pathlib import Path
import configparser
import time
import sys
import re
from logger import get_configured_logger


class SharedStorageMount(object):
    def __init__(self):
        self.logger = get_configured_logger(self.__class__.__name__)
        self.mount_base_path = "/home/shared_storage"
        self.config_file_path = "/.config/rclone/rclone.conf"
        self.parser = self._get_parser()
        self._set_shared_storage()
        self._set_remote_base_folder_path()
        self.task_id = ast.literal_eval(os.environ.get("STARFLOW_INSTANTIATE_PIECE_KWARGS"))["task_id"]
        self.workflow_run_subpath = os.environ.get("STARFLOW_WORKFLOW_RUN_SUBPATH")

        self.shared_storage_function_map = {
            "gcs": self._setup_gcs_shared_storage_config,
            "aws_s3": self._setup_aws_s3_shared_storage_config,
            "local": self._setup_local_shared_storage_config
        }
        self._setup_shared_storage()

    def _setup_local_shared_storage_config(self):
        raise NotImplementedError("Local is not implemented yet")

    def _setup_gcs_shared_storage_config(self):
        raise NotImplementedError("GCS is not implemented yet")

    def _setup_aws_s3_shared_storage_config(self):
        aws_access_key_id = self.workflow_shared_storage_secrets.get("AWS_ACCESS_KEY_ID", None)
        aws_secret_access_key = self.workflow_shared_storage_secrets.get("AWS_SECRET_ACCESS_KEY", None)
        aws_region_name = self.workflow_shared_storage_secrets.get("AWS_REGION_NAME", None)
        if aws_access_key_id is None or aws_secret_access_key is None or aws_region_name is None:
            raise Exception("AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY or AWS_REGION_NAME is not set")

        bucket = self.workflow_shared_storage.get("bucket", None)
        self.parser.set("aws_s3", "env_auth", 'false')
        self.parser.set("aws_s3", "access_key_id", aws_access_key_id)
        self.parser.set("aws_s3", "secret_access_key", aws_secret_access_key)
        self.parser.set("aws_s3", "region", aws_region_name)

        # Update rclone config file with this session details
        with open(self.config_file_path, "w+") as f:
            self.parser.write(f)

        self.rclone_base_path = f"{self.shared_storage_source_name}:{bucket}{self.remote_base_folder_path}{self.workflow_run_subpath}"

    def _setup_shared_storage(self):
        self.shared_storage_function_map[self.shared_storage_source_name]()

    def _set_shared_storage(self):
        self.workflow_shared_storage = ast.literal_eval(os.environ.get("STARFLOW_WORKFLOW_SHARED_STORAGE", {}))
        self.workflow_shared_storage_secrets = ast.literal_eval(os.environ.get("STARFLOW_WORKFLOW_SHARED_STORAGE_SECRETS", {}))
        self.shared_storage_source_name = self.workflow_shared_storage.get("source", None)
        self.shared_storage_mode = self.workflow_shared_storage.get("mode", None)

        if self.shared_storage_source_name is None:
            raise Exception("SHARED_STORAGE_SOURCE_NAME is not set")
        if self.shared_storage_mode is None:
            raise Exception("Shared Storage Mode is not set")

    def _set_remote_base_folder_path(self):
        self.remote_base_folder_path = self.workflow_shared_storage.get("base_folder", "/")
        if self.remote_base_folder_path == "":
            self.remote_base_folder_path = "/"
        if self.remote_base_folder_path[-1] != "/":
            self.remote_base_folder_path += "/"
        if self.remote_base_folder_path[0] != "/":
            self.remote_base_folder_path = "/" + self.remote_base_folder_path

    def _get_parser(self):
        parser = configparser.ConfigParser()
        parser.read(self.config_file_path)
        return parser

    def _mount_upstreams(self, shared_storage_upstream_ids_list: List[str]):
        self.logger.info("Mounting upstreams: %s", shared_storage_upstream_ids_list)
        for upstream_id in shared_storage_upstream_ids_list:
            upstream_mount_dir_path = Path(self.mount_base_path) / upstream_id
            upstream_mount_dir_path.mkdir(parents=True, exist_ok=True)
            subprocess.run([
                "rclone", 
                "mount",  
                f"{self.rclone_base_path}/{upstream_id}", 
                f"{self.mount_base_path}/{upstream_id}", 
                "--allow-other",              # Allow other users to access the mount - might be needed for the other container
                "--allow-non-empty",          # Allow mounting over a non-empty directory
                "--daemon",                   # Run in background
                "--vfs-cache-mode", "full",   # Cache all read and written files on disk
                "--vfs-cache-max-age", "2h",  # Cache files for up to 2 hours
                "--read-only",                # Read only 
                "--config", self.config_file_path  # Use updated config file
            ])

    def _unmount_upstreams(self, shared_storage_upstream_ids_list: List[str]):
        self.logger.info("Unmounting upstreams: %s", shared_storage_upstream_ids_list)
        for upstream_id in shared_storage_upstream_ids_list:
            self.logger.info(f'Checking if mount is busy for upstream {upstream_id}')
            self._check_mount_busy(task_id=upstream_id)

            out = False
            while not out:
                r = subprocess.run(
                    ["fusermount", "-u", f"{self.mount_base_path}/{upstream_id}"], 
                    stdout=subprocess.PIPE, 
                    stderr=subprocess.PIPE, text=True
                )
                self.logger.info(r.stdout)
                if r.stderr:
                    self.logger.error(f"Unmounting failed for upstream {upstream_id}, trying again in 2 seconds")
                    self.logger.error(r.stderr)
                time.sleep(2)
                out = r.returncode == 0
    
    @staticmethod
    def _parse_stdout_rclone_check(stdout_message: str) -> Tuple[Union[int, None], Union[int, None]]:
        matching_files_number = re.search(r': (\d+) matching', stdout_message)
        if matching_files_number:
            matching_files_number = int(matching_files_number.group(1))
        else:
            matching_files_number = 0

        differences_files_number = re.search(r': (\d+) differences', stdout_message)
        if differences_files_number:
            differences_files_number = int(differences_files_number.group(1))
        
        return matching_files_number, differences_files_number

    def _check_mount_busy(self, task_id: str):
        self.logger.info("Checking if mount is busy...")
        source_mount_path = str(Path(self.mount_base_path) / task_id)
        destination_mount_path = str(Path(self.rclone_base_path) / task_id)
        command = ['rclone', 'check', f'{source_mount_path}', f'{destination_mount_path}', '--size-only', '--one-way', "--config", self.config_file_path]
        shared_storage_files_paths = [os.path.join(root, filename) for root, _, filenames in os.walk(source_mount_path) for filename in filenames]

        while True:
            subprocess_output = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            matching_files_number, differences_files_number = self._parse_stdout_rclone_check(subprocess_output.stderr.decode('utf-8'))
            if len(shared_storage_files_paths) == matching_files_number and differences_files_number == 0:
                break
            self.logger.info("Mount is busy, waiting to unmount.")
            time.sleep(2)
        self.logger.info('Mount is ready to unmount.')

    def mount(self):
        shared_storage_upstream_ids_list = ast.literal_eval(os.environ.get("AIRFLOW_UPSTREAM_TASKS_IDS_SHARED_STORAGE", "[]"))
        self._mount_upstreams(shared_storage_upstream_ids_list)

        if self.shared_storage_mode == 'Read':
            # If the shared storage is read only, we don't need to mount the current task. The pod only need access to the upstreams data
            return

        self.logger.info("Mounting task: %s", self.task_id)
        command = [
            "rclone", 
            "mount",  
            f"{self.rclone_base_path}/{self.task_id}", 
            f"{self.mount_base_path}/{self.task_id}", 
            "--allow-other",              # Allow other users to access the mount - might be needed for the other container
            "--allow-non-empty",          # Allow mounting over a non-empty directory
            "--daemon",                   # Run in background       
            "--vfs-cache-mode", "full",   # Cache all read and written files on disk
            "--vfs-cache-max-age", "2h",  # Cache files for up to 2 hours
            "--config", self.config_file_path  # Use updated config file
        ]
        self.logger.info("Mount command: %s", " ".join(command))
        subprocess.run(command, stdout=subprocess.PIPE)
        self.logger.info("Mount finished")
        self.generate_paths(task_id=self.task_id)
        self.logger.info("Mount daemon process running")

    def unmount(self):
        self.logger.info('Unmounting task: %s', self.task_id)
        shared_storage_upstream_ids_list = ast.literal_eval(os.environ.get("AIRFLOW_UPSTREAM_TASKS_IDS_SHARED_STORAGE", "[]"))
        self._unmount_upstreams(shared_storage_upstream_ids_list)

        self._check_mount_busy(task_id=self.task_id)
        out = False
        while not out:
            r = subprocess.run(
                ["fusermount", "-u", f"{self.mount_base_path}/{self.task_id}"], 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE, text=True
            )
            self.logger.info(r.stdout)
            if r.stderr:
                self.logger.error("Unmounting failed, trying again in 2 seconds")
                self.logger.error(r.stderr)
            time.sleep(2)
            out = r.returncode == 0
    
    def generate_paths(self, task_id: str):
        """
        Generate paths to store results, xcom and report data.
            workflow_shared_storage: str - Base path to shared storage
            task_id: str - Task ID
        """
        self.logger.info("Generating paths")
        if not Path(f"{self.mount_base_path}/{task_id}").is_dir():
            Path(f"{self.mount_base_path}/{task_id}").mkdir(parents=True, exist_ok=True)

        # Path to store results data
        results_path = f"{self.mount_base_path}/{task_id}/results"
        if not Path(results_path).is_dir():
            Path(results_path).mkdir(parents=True, exist_ok=True)
        
        # Path to store XCOM data - XCOM data is saved here AND also at /airflow/xcom/return.json
        xcom_path = f"{self.mount_base_path}/{task_id}/xcom"
        if not Path(xcom_path).is_dir():
            Path(xcom_path).mkdir(parents=True, exist_ok=True)
        
        # Path to store report data
        report_path = f"{self.mount_base_path}/{task_id}/report"
        if not Path(report_path).is_dir():
            Path(report_path).mkdir(parents=True, exist_ok=True)


if __name__ == '__main__':
    """
    Execute mount or unmount based on the first argument passed to the script.
    """
    shared_storage_mount = SharedStorageMount()
    getattr(shared_storage_mount, sys.argv[1])()



import os
import pytest

def skip_envs(*envs):
    env = os.environ.get('STARFLOW_TESTS_ENVIRONMENT')
    return pytest.mark.skipif(env in list(envs),reason=f"Not suitable envrionment {env} for current test")


# Maps the storage types to the required list of credentials necessary to use this storage target
storage_types_required_variables = {
    "local": [],
    "s3": ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_REGION_NAME"],
    "gcs": ["", ""],
    "dropbox": [""],
}


from starflow.base_piece import BaseBranchOutputModel
from starflow.scripts.load_piece import load_piece_branch_models_from_path
import tomli
import tomli_w
import json
import jsonschema
import subprocess
import copy
import uuid
import os
import re
import shutil
import time
from pathlib import Path
from rich.console import Console
from typing import Union

from starflow.cli.utils.constants import COLOR_PALETTE
from starflow.client.github_rest_client import GithubRestClient
from starflow.utils import dict_deep_update


console = Console()


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


CONFIG_REQUIRED_FIELDS = {}


def validate_github_token(token: str) -> bool:
    """
    Validate STARFLOW_GITHUB_ACCESS_TOKEN
    By now it is only accepting the ghp token.

    Args:
        token (str): Github access token (ghp)

    Returns:
        bool: True if token is valid, False otherwise.
    """
    regex = r"ghp_[0-9a-zA-Z]{35,40}"
    pattern = re.compile(regex)
    if pattern.match(token):
        return True
    return False


REQUIRED_ENV_VARS_VALIDATORS = {
    "STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN": {
        "depends": lambda arg: arg.get('OPERATORS_REPOSITORY_SOURCE') == 'github',
        "validator_func": validate_github_token
    },
    "STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS": {
        "depends": lambda arg: arg.get('OPERATORS_REPOSITORY_SOURCE') == 'github',
        "validator_func": validate_github_token
    }
}


def set_config_as_env(key: str, value: Union[str, int, float]) -> None:
    """
    Set an ENV variable with the key and value.

    Args:
        key (Union[str, int]): ENV var name
        value (Union[str, int, float]): ENV var value
    """
    key = str(key).strip().upper()
    os.environ[key] = str(value)


def validate_repository_name(name: str) -> bool:
    regex = r'^[A-Za-z0-9_]*$'
    pattern = re.compile(regex)
    if pattern.match(name):
        return True
    return False


def validate_env_vars() -> None:
    """
    Validate user environment variables.
    The accepted variables are:
        - GITHUB_ACCESS_TOKEN: Token to access GitHub API.
    """
    # Set AIRFLOW_UID from host user id
    # https://airflow.apache.org/docs/apache-airflow/stable/start/docker.html#setting-the-right-airflow-user
    # https://airflow.apache.org/docs/apache-airflow/stable/start/docker.html#environment-variables-supported-by-docker-compose
    uid = subprocess.run(["id", "-u"], capture_output=True, text=True)
    set_config_as_env("AIRFLOW_UID", int(uid.stdout))

    for var, validator in REQUIRED_ENV_VARS_VALIDATORS.items():
        if 'depends' in validator:
            should_exist = validator.get('depends')(CONFIG_REQUIRED_FIELDS)
        if not should_exist:
            continue
        env_var = os.environ.get(var, None)
        if env_var:
            continue
        console.print(f"{var} is not defined", style=f"bold {COLOR_PALETTE.get('warning')}")
        new_var = input(f"Enter the {var} value: ")
        while not validator.get('validator_func')(new_var):
            new_var = input(f"Wrong {var} format. Enter a new value: ")
        os.environ[var] = new_var


def validate_config(config_dict: dict) -> None:
    """
    Validate config.toml file.
    """
    required_fields = ["REGISTRY_NAME", "REPOSITORY_NAME", "VERSION"]
    sections = config_dict.keys()
    for section in sections:
        for key, value in config_dict.get(section).items():
            # Check if required fields were defined in config file
            if key in required_fields:
                required_fields.remove(key)
                # Set config vars as env vars, it will be used by docker build
                set_config_as_env(key, value)
    if len(required_fields) > 0:
        console.print("Missing required fields: {}".format(required_fields), style=f"bold {COLOR_PALETTE.get('error')}")


def validate_repository_structure() -> None:
    """
    Validate the Pieces repository structure.
    The basic initial structure must contain:
    - config.toml
    - pieces/
    - dependencies/
    """

    organized_starflow_path = Path(".") / ".starflow/"
    if not organized_starflow_path.is_dir():
        organized_starflow_path.mkdir(parents=True, exist_ok=True)

    # Validating config
    config_path = Path(".") / 'config.toml'
    if not config_path.is_file():
        console.print("Missing config.toml file", style=f"bold {COLOR_PALETTE.get('error')}")
        raise FileNotFoundError("Missing config.toml file")

    with open(config_path, "rb") as f:
        config_dict = tomli.load(f)

    validate_config(config_dict)
    validate_env_vars()

    pieces_repository = Path(".")
    if not pieces_repository.is_dir():
        console.print("Pieces repository path does not exist", style=f"bold {COLOR_PALETTE.get('error')}")
        raise Exception("Pieces repository path does not exist")

    if not (pieces_repository / 'pieces').is_dir():
        console.print("Pieces directory does not exist", style=f"bold {COLOR_PALETTE.get('error')}")
        raise Exception("Pieces directory does not exist")

    if not (pieces_repository / 'dependencies').is_dir():
        console.print("Dependencies directory does not exist", style=f"bold {COLOR_PALETTE.get('error')}")
        raise Exception("Dependencies directory does not exist")

    # os.environ['OPERATORS_REPOSITORY_PATH_HOST'] = str(pieces_repository)


def validate_pieces_folders() -> None:
    """
    Validate the Pieces folders from an Pieces repository.
    """
    from starflow.schemas import PieceMetadata

    pieces_path = Path(".") / "pieces"
    dependencies_path = Path(".") / "dependencies"
    organized_starflow_path = Path(".") / ".starflow/"
    dependencies_files = [f.name for f in dependencies_path.glob("*")]
    name_errors = list()
    missing_file_errors = list()
    missing_dependencies_errors = list()
    for op_dir in pieces_path.glob("*Piece"):
        if op_dir.is_dir():
            # Validate necessary files exist
            files_names = [f.name for f in op_dir.glob("*")]
            if 'models.py' not in files_names:
                missing_file_errors.append(f"missing 'models.py' for {op_dir.name}")
            if 'piece.py' not in files_names:
                missing_file_errors.append(f"missing 'piece.py' for {op_dir.name}")
            if len(missing_file_errors) > 0:
                raise Exception('\n'.join(missing_file_errors))

            # Validate metadata
            if (op_dir / "metadata.json").is_file():
                with open(str(op_dir / "metadata.json"), "r") as f:
                    metadata = json.load(f)
                jsonschema.validate(instance=metadata, schema=PieceMetadata.model_json_schema())

                # Validate Pieces name
                if metadata.get("name", None) and not metadata["name"] == op_dir.name:
                    name_errors.append(op_dir.name)

                # Validate dependencies exist
                if metadata.get("dependency", None):
                    req_file = metadata["dependency"].get("requirements_file", None)
                    if req_file and req_file != "default" and req_file not in dependencies_files:
                        missing_dependencies_errors.append(f'missing dependency file {req_file} defined for {op_dir.name}')

                    dock_file = metadata["dependency"].get("dockerfile", None)
                    if dock_file and dock_file != "default" and dock_file not in dependencies_files:
                        missing_dependencies_errors.append(f'missing dependency file {dock_file} defined for {op_dir.name}')

    if len(name_errors) > 0:
        raise Exception(f"The following Pieces have inconsistent names: {', '.join(name_errors)}")
    if len(missing_dependencies_errors) > 0:
        raise Exception("\n" + "\n".join(missing_dependencies_errors))


def create_pieces_repository(repository_name: str, container_registry: str) -> None:
    """
    Create a new Pieces repository from template, with folder structure and placeholder files.
    """
    while not validate_repository_name(repository_name):
        repository_name = input("\nInvalid repository name. Should have only numbers, letters and underscores. \nEnter a new repository name: ") or f"new_repository_{str(uuid.uuid4())[0:8]}"
    cwd = Path.cwd()
    repository_folder = cwd / repository_name
    if repository_folder.is_dir():
        raise Exception("Repository folder already exists")
    console.print(f"Cloning template Pieces repository at: {repository_folder}")
    subprocess.run(["git", "clone", "https://github.com/prochain-star-atlas/starflow_pieces_repository_template.git", repository_name], capture_output=True, text=True)
    shutil.rmtree(f"{repository_name}/.git")

    # Update config
    with open(f"{repository_name}/config.toml", "rb") as f:
        repo_config = tomli.load(f)

    repo_config["repository"]["REPOSITORY_NAME"] = repository_name
    repo_config["repository"]["REGISTRY_NAME"] = container_registry if container_registry else "enter-your-github-registry-name-here"

    with open(f"{repository_name}/config.toml", "wb") as f:
        tomli_w.dump(repo_config, f)

    console.print(f"Pieces repository successfully create at: {repository_folder}", style=f"bold {COLOR_PALETTE.get('success')}")
    console.print("")


def create_compiled_pieces_metadata(source_url: str | None = None) -> None:
    """
    Create compiled metadata from Pieces metadata.json files and include input_schema generated from models.py
    """
    from starflow.scripts.load_piece import load_piece_models_from_path
    from starflow.utils.metadata_default import metadata_default

    pieces_path = Path(".") / "pieces"
    compiled_metadata = dict()
    for op_dir in pieces_path.glob("*Piece"):
        if op_dir.is_dir():
            piece_name = op_dir.name

            # Update with user-defined metadata.json
            metadata = copy.deepcopy(metadata_default)
            if (op_dir / "metadata.json").is_file():
                with open(str(op_dir / "metadata.json"), "r") as f:
                    metadata_op = json.load(f)
                metadata_op['name'] = metadata_op.get('name', piece_name)
                dict_deep_update(metadata, metadata_op)
            else:
                metadata['name'] = piece_name

            # Add input and output schemas

            if metadata["branchPiece"] == True:

                input_model_class = load_piece_branch_models_from_path(
                    pieces_folder_path=str(pieces_path),
                    piece_name=op_dir.name
                )
                metadata["input_schema"] = input_model_class.model_json_schema()
                metadata["output_schema"] = BaseBranchOutputModel.model_json_schema()
                metadata["secrets_schema"] = None

            else:

                input_model_class, output_model_class, secrets_model_class = load_piece_models_from_path(
                    pieces_folder_path=str(pieces_path),
                    piece_name=op_dir.name
                )
                metadata["input_schema"] = input_model_class.model_json_schema()
                metadata["output_schema"] = output_model_class.model_json_schema()
                metadata["secrets_schema"] = secrets_model_class.model_json_schema() if secrets_model_class else None

            # Add source code url
            metadata["source_url"] = None
            if source_url and len(source_url) > 0:
                metadata["source_url"] = source_url + f"/tree/main/pieces/{piece_name}"

            # Add to compiled metadata
            compiled_metadata[piece_name] = metadata

    # Save compiled_metadata.json file
    organized_starflow_path = Path(".") / ".starflow/"
    with open(str(organized_starflow_path / "compiled_metadata.json"), "w") as f:
        json.dump(compiled_metadata, f, indent=4)


def create_dependencies_map(save_map_as_file: bool = True) -> None:
    """
    Construct a map between Pieces and unique definitions for docker images dependencies

    Args:
        save_map_as_file (bool, optional): Set if dependencies_map will be saved as file. Defaults to True.

    Raises:
        ValueError: Raise if pieces is not found in the pieces_repository
    """
    organized_starflow_path = Path(".") / ".starflow/"
    with open(organized_starflow_path / "compiled_metadata.json", "r") as f:
        compiled_metadata = json.load(f)

    pieces_images_map = {}
    for op_i, (piece_name, piece_metadata) in enumerate(compiled_metadata.items()):

        if piece_metadata.get("secrets_schema"):
            piece_secrets = set(piece_metadata.get("secrets_schema")["properties"].keys())
        else:
            piece_secrets = set()

        if op_i == 0:
            pieces_images_map = {
                f"group0": {
                    "dependency": piece_metadata["dependency"],
                    "pieces": [piece_name],
                    "secrets": piece_secrets
                }
            }
        else:
            # Compare with metadata from previous pieces to see if a new docker image needs to be built
            existing_keys = pieces_images_map.keys()
            skip_new_image = False
            for i, dep_key in enumerate(existing_keys):
                if all([piece_metadata["dependency"][k] == pieces_images_map[dep_key]["dependency"][k] for k in piece_metadata["dependency"].keys()]):
                    pieces_images_map[dep_key]["pieces"].append(piece_name)
                    pieces_images_map[dep_key]["secrets"].update(piece_secrets)
                    skip_new_image = True
                    continue
            if not skip_new_image:
                pieces_images_map[f"group{len(existing_keys)}"] = {
                    "dependency": piece_metadata["dependency"],
                    "pieces": [piece_name],
                    "secrets": piece_secrets
                }

    if not pieces_images_map:
        raise ValueError("No pieces found in the Pieces Repository")

    if save_map_as_file:
        map_file_path = organized_starflow_path / "dependencies_map.json"
        with open(map_file_path, "w") as outfile:
            json.dump(pieces_images_map, outfile, indent=4, cls=SetEncoder)


def build_docker_images(tag_overwrite: str | None = None, dev: bool = False) -> None:
    """
    Convenience function to build Docker images from the repository dependencies and publish them to Docker Hub
    """
    from starflow.scripts.build_docker_images_pieces import build_images_from_pieces_repository

    console.print("Building Docker images and generating map file...")
    updated_dependencies_map = build_images_from_pieces_repository(tag_overwrite=tag_overwrite, dev=dev)
    return updated_dependencies_map


def publish_docker_images() -> None:
    """
    Load pieces to docker image map from environment variable and publish them to Container Registry.
    This is used in the CI/CD pipeline to publish images to Container Registry after they are built and the tests pass.
    """
    from starflow.scripts.build_docker_images_pieces import publish_image

    pieces_images_map = json.loads(os.environ.get('PIECES_IMAGES_MAP', '{}'))
    if not pieces_images_map:
        raise ValueError("No images found to publish.")

    console.print("Publishing Docker images...")
    all_images = set([e for e in pieces_images_map.values()])
    for image in all_images:
        console.print(f"Publishing image {image}...")
        publish_image(source_image_name=image)


def validate_repo_name(repo_name: str) -> None:
    """
    Validate repository name
    """
    if any([a.isspace() for a in repo_name]):
        raise ValueError("Repository name should not contain blank spaces")
    if any([a in repo_name for a in ["!", "@", "#", "$", "%", "^", "&", "*", "(", ")", "+", "=", "{", "}", "[", "]", ":", ";", "'", '"', "<", ">", "?", "/", "\\", "|", "~", "`"]]):
        raise ValueError("Repository name should not contain special characters")


def organize_pieces_repository(
    build_images: bool,
    source_url: str,
    tag_overwrite: str | None = None,
    dev: bool = False
) -> None:
    """
    Organize Piece's repository for starflow. This will:
    - validate the folder structure, and create the pieces compiled_metadata.json and dependencies_map.json files
    - build Docker images for the Pieces
    - publish images at github container registry
    """
    # Validate repository
    console.print("Validating repository structure and files...")
    validate_repository_structure()
    validate_pieces_folders()
    console.print("Validation successful!", style=f"bold {COLOR_PALETTE.get('success')}")

    # Load config
    with open("config.toml", "rb") as f:
        repo_config = tomli.load(f)

    # Validate repository name
    repo_name = repo_config["repository"]["REPOSITORY_NAME"]
    validate_repo_name(repo_name)

    # Tag overwrite
    if tag_overwrite:
        repo_config["repository"]["VERSION"] = tag_overwrite

    # Create compiled metadata from Pieces metadata.json files and add data input schema
    create_compiled_pieces_metadata(source_url=source_url)

    # Generate dependencies_map.json file
    create_dependencies_map(save_map_as_file=True)
    console.print("Metadata and dependencies organized successfully!", style=f"bold {COLOR_PALETTE.get('success')}")

    # Build and publish the images
    if build_images:
        updated_dependencies_map = build_docker_images(tag_overwrite=tag_overwrite, dev=dev)
        map_file_path = Path(".") / ".starflow/dependencies_map.json"
        with open(map_file_path, "w") as outfile:
            json.dump(updated_dependencies_map, outfile, indent=4)


def create_release(tag_name: str | None = None, commit_sha: str | None = None):
    """
    Create a new release and tag in the repository for the latest commit.
    """
    token = os.environ.get('GITHUB_TOKEN', None)
    if not token:
        raise ValueError("GITHUB_TOKEN not found in ENV vars.")
    client = GithubRestClient(token=token)

    # Get version from config.toml
    with open("config.toml", "rb") as f:
        repo_config = tomli.load(f)

    version = repo_config.get("repository", {}).get("VERSION", None)
    if not version:
        raise ValueError("VERSION not found in config.toml")

    # Overwrite version if tag_name is passed
    if tag_name:
        version = tag_name

    # https://docs.github.com/en/actions/learn-github-actions/contexts#example-printing-context-information-to-the-log
    # Passing from context to env - ${{ github.repository }} - get repository from context
    repository = os.environ.get('GITHUB_REPOSITORY', None)
    if not repository:
        raise ValueError("GITHUB_REPOSITORY not found in ENV vars.")

    # Check if tag already exists
    tag = client.get_tag(repo_name=repository, tag_name=version)
    if tag:
        raise ValueError(f'Tag {version} already exists')

    # Get latest commit
    if not commit_sha:
        latest_commit = client.get_commits(repo_name=repository, number_of_commits=1)[0]
        commit_sha = latest_commit.sha
    if not commit_sha:
        raise ValueError("Commit SHA not found")

    # Create tag
    release = client.create_release(
        repo_name=repository,
        version=version,
        tag_message=f'Release {version}',
        release_message=f'Release {version}',
        target_commit_sha=commit_sha,
    )
    console.print(f"Release {version} created successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
    return release


def delete_release(tag_name: str):
    """
    Delete a release with given tag from the repository.
    """
    token = os.environ.get('GITHUB_TOKEN', None)
    if not token:
        raise ValueError("GITHUB_TOKEN not found in ENV vars.")
    client = GithubRestClient(token=token)

    # https://docs.github.com/en/actions/learn-github-actions/contexts#example-printing-context-information-to-the-log
    # Passing from context to env - ${{ github.repository }} - get repository from context
    repository = os.environ.get('GITHUB_REPOSITORY', None)
    if not repository:
        raise ValueError("GITHUB_REPOSITORY not found in ENV vars.")

    # Check if tag already exists
    tag = client.get_tag(repo_name=repository, tag_name=tag_name)
    if tag:
        # Delete release by tag
        client.delete_release_by_tag(repo_name=repository, tag_name=tag_name)
        client.delete_tag(repo_name=repository, tag_name=tag_name)
        console.print(f"Attempting to delete release {tag_name}...", style="bold")
        timeout = 30  # 30 seconds timeout
        start_time = time.time()
        while time.time() - start_time < timeout:
            if not client.get_tag(repo_name=repository, tag_name=tag_name):
                console.print(f"Release {tag_name} deleted successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                return
            time.sleep(5)  # Wait for 5 seconds before checking again
        console.print(f"Deletion error: Release {tag_name} still exists after {timeout} seconds.", style=f"bold {COLOR_PALETTE.get('warning')}")
    else:
        console.print(f"Release {tag_name} not found. Skipping deletion.", style=f"bold {COLOR_PALETTE.get('warning')}")



import abc
import json
import os
import ast
from datetime import datetime
from pathlib import Path
from typing import Union
import pydantic
import pickle
import time
import subprocess
import base64
from typing import Optional

from starflow.logger import get_configured_logger
from starflow.schemas import DeployModeType, DisplayResultFileType
from starflow.exceptions.exceptions import InvalidPieceOutputError

from pydantic import BaseModel, Field

class BaseBranchOutputModel(BaseModel):
    """
    Branch Piece Output Model
    """

    branch_main: bool = Field(
        default=True,
        description="Main branch selected",
    )

class BasePiece(metaclass=abc.ABCMeta):

    @classmethod
    def set_metadata(cls, metadata):
        """
        _summary_

        Args:
            metadata (_type_): _description_
        """
        # Piece name as used by Airflow
        cls.__name__ = metadata.get("name", "BasePiece")

        # Full metadata
        cls._metadata_ = metadata

    def __init__(
        self,
        deploy_mode: DeployModeType,
        task_id: str,
        dag_id: str,
    ) -> None:
        """
        The base class from which every starflow custom Piece should inherit from.
        BasePiece methods and variables that can be used by inheriting Pieces classes:
        self.results_path - Path to store results data
        self.logger - Logger functionality

        Args:
            deploy_mode (DeployModeType): _description_
            task_id (str): _description_
            dag_id (str): _description_
        """

        # Piece task attributes
        self.task_id = task_id
        self.dag_id = dag_id
        self.deploy_mode = "k8s"

        # Logger
        self.logger = get_configured_logger(f"{self.__class__.__name__ }-{self.task_id}")

        self.display_result = None
        self._shared_storage_usage_in_bytes = 0

    def start_logger(self):
        """
        Start logger.
        """
        self.logger.info(f"Started {self.task_id} of type {self.__class__.__name__} at {str(datetime.now().isoformat())}")
        self.logger.info("Start cut point for logger 48c94577-0225-4c3f-87c0-8add3f4e6d4b")

    def _wait_for_sidecar_paths(self):
        # Wait for sidecar create directories
        while True:
            if Path(self.report_path).is_dir():
                break
            time.sleep(2)

    @staticmethod
    def _get_folder_size(folder_path):
        total_size = 0
        for dirpath, _, filenames in os.walk(folder_path):
            for filename in filenames:
                file_path = os.path.join(dirpath, filename)
                total_size += os.path.getsize(file_path)
        return total_size

    def generate_paths(self):
        """
        Generates paths for shared storage.
        """
        # Base path for fetching and storing runs results
        if not Path(self.workflow_shared_storage_path).is_dir():
            Path(self.workflow_shared_storage_path).mkdir(parents=True, exist_ok=True)

        # # Path to store results data
        if not Path(self.results_path).is_dir():
            Path(self.results_path).mkdir(parents=True, exist_ok=True)

        # # Path to store XCOM data - XCOM data is saved here AND also at /airflow/xcom/return.json
        if not Path(self.xcom_path).is_dir():
            Path(self.xcom_path).mkdir(parents=True, exist_ok=True)

        # # Path to store report data
        if not Path(self.report_path).is_dir():
            Path(self.report_path).mkdir(parents=True, exist_ok=True)

    def get_upstream_tasks_data(self):
        """
        Get XCOM and results data from upstream tasks. Stores this information in the attribute `self.upstream_task_data` as a dictionary with the following structure:
        {
            "task_id_1": {
                "xcom": {...},
                "results": "path/to/results"
            },
            "task_id_2": {...},
        }

        Raises:
            NotImplementedError: _description_
        """
        self.upstream_tasks_data = dict()
        upstream_tasks_ids = ast.literal_eval(os.getenv("AIRFLOW_UPSTREAM_TASKS_IDS", str(list())))
        for tid in upstream_tasks_ids:
            self.upstream_tasks_data[tid] = dict()
            self.upstream_tasks_data[tid]["results"] = self.workflow_shared_storage_path + f"/{tid}/results"
            with open(f"{self.workflow_shared_storage_path}/{tid}/xcom/return.json") as f:
                self.upstream_tasks_data[tid]["xcom"] = json.load(f)

    def validate_and_get_env_secrets(self, piece_secrets_model: pydantic.BaseModel = None):
        """
        Get secret variables for this Piece from ENV. The necessary secret variables to run the Piece should be defined in the Piece's SecretsModel.
        The secrets can then be retrieved and used in the Piece's `piece_function` method.

        Args:
            piece_secrets_model (pydantic.BaseModel): _description_
        """
        if piece_secrets_model:
            secrets_values = ast.literal_eval(os.environ.get('STARFLOW_PIECE_SECRETS', '{}'))
            return piece_secrets_model(**secrets_values)
        return None

    def format_xcom(self, output_obj: pydantic.BaseModel) -> dict:
        """
        Formats and adds extra metadata to XCOM dictionary content.

        Args:
            output_obj (dict): Pydantic output model

        Returns:
            dict: XCOM dictionary
        """
        # xcom_obj = output_obj.model_dump()
        xcom_obj = json.loads(output_obj.model_dump_json())
        if not isinstance(xcom_obj, dict):
            self.logger.info(f"Piece {self.__class__.__name__} is not returning a valid XCOM object. Auto-generating a base XCOM for it...")
            xcom_obj = dict()

        # Serialize self.display_result and add it to XCOM
        if isinstance(self.display_result, dict):
            if "file_type" not in self.display_result:
                raise Exception("display_result must have 'file_type' key")
            if "base64_content" not in self.display_result:
                if "file_path" not in self.display_result:
                    raise Exception("self.display_result dict must have either 'file_path' or 'base64_content' keys")
                self.display_result["base64_content"] = self.serialize_display_result_file(
                    file_path=self.display_result["file_path"],
                    file_type=self.display_result["file_type"]
                )
            self.display_result["file_path"] = str(self.display_result.get("file_path", None))
            self.display_result["file_type"] = str(self.display_result["file_type"])
        else:
            raw_content = f"Piece {self.__class__.__name__} did not return a valid display_result."
            base64_content = base64.b64encode(raw_content.encode("utf-8")).decode("utf-8")
            self.display_result = dict()
            self.display_result["file_path"] = None
            self.display_result["file_type"] = "txt"
            self.display_result["base64_content"] = base64_content

        xcom_obj["display_result"] = self.display_result

        # Update XCOM with extra metadata
        xcom_obj.update(
            piece_name=self.__class__.__name__,
            piece_metadata=self._metadata_
        )
        return xcom_obj

    def push_xcom(self, xcom_obj: dict):
        """
        Push piece's output to XCOM, to be used by downstream pieces.

        Args:
            xcom_obj (dict): Formatted XCOM object as a dictionary

        Raises:
            NotImplementedError: _description_
        """

        # In Kubernetes, return XCom must be stored in /airflow/xcom/return.json
        # https://airflow.apache.org/docs/apache-airflow-providers-cncf-kubernetes/stable/pieces.html#how-does-xcom-work
        file_path = Path('/airflow/xcom/return.json')
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(file_path), 'w') as fp:
            json.dump(xcom_obj, fp)

        # Also store it at /home/workflow_shared_data/{self.task_id}/xcom/return.json for convenience
        file_path = self.xcom_path + "/return.json"
        with open(file_path, 'w') as fp:
            json.dump(xcom_obj, fp)


    def run_piece_function(
        self,
        piece_input_data: dict,
        piece_input_model: pydantic.BaseModel,
        piece_output_model: pydantic.BaseModel,
        piece_secrets_model: Optional[pydantic.BaseModel] = None,
        airflow_context: Optional[dict] = None,
        workspace_id: Optional[int] = None
    ):
        """
        _summary_

        Args:
            piece_input_data (dict): Dictionary containing Piece's Input kwargs
            piece_input_model (pydantic.BaseModel): Piece's InputModel
            piece_output_model (pydantic.BaseModel): Piece's OutputModel
            piece_secrets_model (pydantic.BaseModel, optional): Piece's SecretsModel. Defaults to None.
            airflow_context (dict, optional): Airflow context dictionary. Defaults to None.

        Raises:
            InvalidPieceOutputError: _description_
        """
        # Start logger
        self.start_logger()

        self.piece_input_model = piece_input_model
        self.piece_output_model = piece_output_model
        self.piece_secrets_model = piece_secrets_model

        # Airflow context dictionary: https://composed.blog/airflow/execute-context
        # For local-bash and kubernetes deploy modes, we assemble this ourselves and the context data is more limited
        if airflow_context is None:
            self.airflow_context = {
                "execution_datetime": os.getenv('AIRFLOW_CONTEXT_EXECUTION_DATETIME', "123456789"),
                "dag_run_id": os.getenv('AIRFLOW_CONTEXT_DAG_RUN_ID', "123456789"),
            }

        # Check if Piece's necessary secrets are present in ENV
        secrets_model_obj = self.validate_and_get_env_secrets(piece_secrets_model=piece_secrets_model)

        # Generate paths
        workflow_run_subpath = os.environ.get('STARFLOW_WORKFLOW_RUN_SUBPATH', '')
        self.workflow_shared_storage_path = Path("/home/shared_storage")
        if self.deploy_mode == 'local-compose':
            self.workflow_shared_storage_path = str(self.workflow_shared_storage_path / workflow_run_subpath)
        self.results_path = f"{self.workflow_shared_storage_path}/{self.task_id}/results"
        self.xcom_path = f"{self.workflow_shared_storage_path}/{self.task_id}/xcom"
        self.report_path = f"{self.workflow_shared_storage_path}/{self.task_id}/report"
        shared_storage_source_name = os.environ.get('STARFLOW_WORKFLOW_SHARED_STORAGE_SOURCE_NAME', None)
        if not shared_storage_source_name or shared_storage_source_name == "none" or self.deploy_mode == "local-compose":
            self.generate_paths()
        else:
            self._wait_for_sidecar_paths()

        # Using pydantic to validate input data
        input_model_obj = piece_input_model(**piece_input_data)

        # Run piece function
        call_piece_func_dict = {"input_data": input_model_obj}
        if piece_secrets_model:
            call_piece_func_dict['secrets_data'] = secrets_model_obj
        call_piece_func_dict['workspace_id'] = workspace_id
        output_obj = self.piece_function(**call_piece_func_dict)

        # Validate output data
        if isinstance(output_obj, dict):
            output_obj = piece_output_model(**output_obj)
        if not isinstance(output_obj, piece_output_model):
            raise InvalidPieceOutputError(piece_name=self.__class__.__name__)

        # Push XCom
        xcom_obj = self.format_xcom(output_obj=output_obj)
        shared_storage_base_path = f"{self.workflow_shared_storage_path}/{self.task_id}"
        self._shared_storage_usage_in_bytes = self._get_folder_size(shared_storage_base_path)
        xcom_obj['_shared_storage_usage_in_bytes'] = self._shared_storage_usage_in_bytes
        self.push_xcom(xcom_obj=xcom_obj)
        self.logger.info(f"Piece used {self._shared_storage_usage_in_bytes} bytes of storage.")
        self.logger.info("End cut point for logger 48c94577-0225-4c3f-87c0-8add3f4e6d4b")


    @classmethod
    def dry_run(
        cls,
        input_data: dict,
        piece_input_model: pydantic.BaseModel,
        piece_output_model: pydantic.BaseModel,
        piece_secrets_model: pydantic.BaseModel = None,
        secrets_data: dict = None,
        results_path: str = None,
    ):
        # Instantiate models
        input_model_obj = piece_input_model(**input_data)
        secrets_model_obj = piece_secrets_model(**secrets_data) if piece_secrets_model else None

        class DryPiece(cls):
            def __init__(self, results_path):
                self.results_path = results_path
                self.logger = get_configured_logger(f"{self.__class__.__name__ }-dry-run")

        dry_instance = DryPiece(results_path=results_path)

        # Run piece function
        call_piece_func_dict = {
            "self": dry_instance,
            "input_data": input_model_obj
        }
        if piece_secrets_model:
            call_piece_func_dict['secrets_data'] = secrets_model_obj
        output_obj = cls.piece_function(**call_piece_func_dict)

        # Validate output data
        if isinstance(output_obj, dict):
            output_obj = piece_output_model(**output_obj)
        if not isinstance(output_obj, piece_output_model):
            raise InvalidPieceOutputError(piece_name=cls.__name__)

        return output_obj

    @staticmethod
    def get_container_cpu_limit() -> float:
        """
        Get the CPU limit of the container in millicores.
        reference: https://stackoverflow.com/questions/65551215/get-docker-cpu-memory-limit-inside-container/65554131#65554131
        """
        with open("/sys/fs/cgroup/cpu/cpu.cfs_quota_us") as fp:
            cfs_quota_us = int(fp.read())
        with open("/sys/fs/cgroup/cpu/cpu.cfs_period_us") as fp:
            cfs_period_us = int(fp.read())

        container_cpus = float(cfs_quota_us / cfs_period_us)
        return container_cpus

    @staticmethod
    def get_nvidia_smi_output() -> str:
        """
        Get the output of nvidia-smi command.
        """
        try:
            nvidia_smi_output = subprocess.check_output(["nvidia-smi", "-q"])
            return nvidia_smi_output
        except Exception as e:
            raise Exception(f"Error while running nvidia-smi: {e}")

    @staticmethod
    def get_container_memory_limit() -> int:
        """
        Get the memory limit of the container in bytes.
        """
        with open("/sys/fs/cgroup/memory/memory.limit_in_bytes") as fp:
            container_memory_limit = int(fp.read())
        return container_memory_limit

    @staticmethod
    def get_container_memory_usage() -> int:
        """
        Get the memory usage of the container in bytes.
        """
        with open("/sys/fs/cgroup/memory/memory.usage_in_bytes") as fp:
            container_memory_usage = int(fp.read())
        return container_memory_usage

    @abc.abstractmethod
    def piece_function(self):
        """
        This function carries the relevant code for the Piece run.
        It should have all the necessary content for auto-generating json schemas.
        All arguments should be type annotated and docstring should carry description for each argument.
        """
        raise NotImplementedError("This method must be implemented in the child class!")

    def serialize_display_result_file(self, file_path: Union[str, Path], file_type: DisplayResultFileType) -> dict:
        """
        Serializes the content of 'display_result_file' into base64 string, to fit Airflow XCOM.

        Args:
            file_path (Union[str, Path]): The path to the file.
            file_type (DisplayResultFileType): The type of the file.

        Returns:
            dict: A dictionary containing the base64-encoded content and the file type.
        """
        if not Path(file_path).exists():
            self.logger.info(f"File {file_path} does not exist. Skipping serialization...")
            return None
        if not Path(file_path).is_file():
            self.logger.info(f"Path {file_path} is not a file. Skipping serialization...")
            return None
        # Read file content as bytes and encode content into base64
        with open(file_path, "rb") as f:
            content_bytes = f.read()
        encoded_content = base64.b64encode(content_bytes).decode('utf-8')
        return encoded_content

    # @abc.abstractmethod
    # def generate_report(self):
    #     """This function carries the relevant code for the Piece report."""
    #     raise NotImplementedError("This method must be implemented in the child class!")



import boto3
import os
from botocore.exceptions import ClientError
from starflow.logger import get_configured_logger


class S3Client(object):
    def __init__(self, bucket_name):
        """
        Credentials & configuration are handled in .aws directory or environment variables
        """
        self.s3r = boto3.resource('s3')
        self.logger = get_configured_logger(self.__class__.__name__)
        self.set_bucket(bucket_name)

    def set_bucket(self, bucket_name):
        """
        boto3.resource("s3").Bucket reference:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#bucket
        """
        try:
            self.bucket = self.s3r.Bucket(bucket_name)
        except ClientError as e:
            self.logger.exception(e)

    def download_file_from_s3_folder(self, object_key: str, file_path: str):
        """ref: https://boto3.amazonaws.com/v1/documentation/api/1.9.42/reference/services/s3.html#S3.Bucket.download_file"""
        self.bucket.download_file(
            Key=object_key, 
            Filename=file_path
        )
    
    def download_file_from_s3_folder_as_bytes(self, object_key: str):
        """ref: https://stackoverflow.com/a/48696641/11483674"""
        s3_object = self.bucket.Object(object_key)
        return s3_object.get()["Body"]

    def download_s3_folder(self, s3_folder, local_dir=None):
        """
        Download the contents of a S3 folder, with sub-directories
        ref: https://stackoverflow.com/a/62945526/11483674
        
        Args:
            bucket_name: the name of the s3 bucket
            s3_folder: the folder path in the s3 bucket
            local_dir: a relative or absolute directory path in the local file system
        """
        try:
            for obj in self.bucket.objects.filter(Prefix=s3_folder):
                target = obj.key if local_dir is None \
                    else os.path.join(local_dir, os.path.relpath(obj.key, s3_folder))
                if not os.path.exists(os.path.dirname(target)):
                    os.makedirs(os.path.dirname(target))
                if obj.key[-1] == '/':
                    continue
                self.bucket.download_file(obj.key, target)
            return None
        except ClientError as e:
            self.logger.exception(e)
            return str(e)
    
    def list_objects_from_bucket(self):
        return [o for o in self.bucket.objects.filter()]

    def upload_file_to_s3(self, object_key: str, file_path: str):
        """ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Bucket.upload_file"""
        self.bucket.upload_file(
            Filename=file_path,
            Key=object_key
        )


from bottle import run, Bottle, request
from starflow.testing.dry_run import piece_dry_run
import json


app = Bottle()


@app.route('/test', method='POST')
def run_test():
    dry_run_args = json.loads(request.body.read())
    output_data = piece_dry_run(**dry_run_args)
    return output_data


@app.route("/health-check")
def health_check():
    return {"status": "ok"}


def run_server():
    run(app=app, host='0.0.0.0', port=8080)



from typing import Any
from starflow.logger import get_configured_logger
from datetime import datetime
import requests
from urllib.parse import urljoin
import os
from keycloak import KeycloakOpenID


class starflowBackendRestClient(requests.Session):
    def __init__(self, base_url: str = None, *args, **kwargs):
        super(starflowBackendRestClient, self).__init__(*args, **kwargs)
        self.logger = get_configured_logger(self.__class__.__name__)
        self.keycloak_openid = KeycloakOpenID(server_url=os.environ["OPEN_ID_SERVER_URL"],
                            client_id=os.environ["OPEN_ID_CLIENT_ID"],
                            realm_name=os.environ["OPEN_ID_REALM_NAME"],
                            client_secret_key=os.environ["OPEN_ID_CLIENT_SECRET"],
                            verify=False)
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = os.environ.get("STARFLOW_BACKEND_HOST", "http://localhost:8000")

    def openid_get_token(self) -> Any:
        token = self.keycloak_openid.token(username=os.environ["OPEN_ID_USERNAME_SERVICE_USER"], password=os.environ["OPEN_ID_PASSWORD_SERVICE_USER"])
        return token

    def request(self, method, resource, **kwargs) -> requests.Response:
        try:
            bearer_token = self.openid_get_token()
            headers = {"Authorization": "Bearer " + bearer_token['access_token']}
            url = urljoin(self.base_url, resource)
            return super(starflowBackendRestClient, self).request(method, url, verify=False, headers=headers, **kwargs)
        except Exception as e:
            self.logger.exception(e)
            raise e

    def health_check(self) -> requests.Response:
        resource = "/health-check"
        response = self.request(
            method="get",
            resource=resource
        )
        return response

    def get_piece_secrets(self, piece_repository_id: int, piece_name: str) -> requests.Response:
        resource = f"/pieces-repositories/{piece_repository_id}/secrets/{piece_name}/secrets-values"
        response = self.request(
            method='get',
            resource=resource
        )
        return response

    def get_piece_repositories_from_workspace_id(self, params: dict) -> requests.Response:
        resource = "/pieces-repositories/worker"
        response = self.request(
            method='get',
            resource=resource,
            params=params
        )
        return response

    def check_create_airflow_connection(self, conn_id: str, conn_type: str):
        """
        This should check if a specific Airflow connection exists and create it if it doesn't.
        ref: https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/post_connection
        """
        # TODO - this is WIP
        resource = f"/airflow/connections/{conn_id}"
        response = self.request(
            method='get',
            resource=resource,
        )
        if response.status_code == 404:
            response = self.request(
                method='post',
                resource=resource,
                json={
                    "conn_id": conn_id,
                    "conn_type": conn_type,
                    "login": "",  # TODO: add login and password
                    "password": "",
                    "extra": {
                        # Specify extra parameters here, e.g. "region_name": "eu-central-1"
                    },
                }
            )
            if response.status_code != 200:
                raise Exception(f"Failed to create Airflow connection {conn_id}. \n {response.json()}")



from airflow import DAG
from airflow.models import BaseOperator
from datetime import datetime
from typing import Callable, Iterable, Optional
import os

from kubernetes.client import models as k8s
from starflow.custom_operators.k8s_branch_operator import starflowKubernetesBranchPodOperator
from starflow.custom_operators.docker_operator import starflowDockerOperator
from starflow.custom_operators.python_operator import PythonOperator
from starflow.custom_operators.worker_operator import starflowWorkerOperator
from starflow.logger import get_configured_logger
from starflow.schemas import shared_storage_map, StorageSource


class TaskBranch(object):
    """
    The TaskBranch object represents a task in a workflow.
    It is only instantiated by processes parsing dag files in Airflow.
    """
    def __init__(
        self,
        dag: DAG,
        task_id: str,
        workspace_id: int,
        piece: dict,
        piece_input_kwargs: dict,
        task_id_main_branch: str | Iterable[str],
        task_id_else_branch: str | Iterable[str],
        workflow_shared_storage: dict = None,
        container_resources: dict = None,
        username: Optional[str] = None,
        trigger_rule="all_success",
        **kwargs
    ) -> None:
        # Task configuration and attributes
        self.task_id = task_id
        self.workspace_id = workspace_id
        self.logger = get_configured_logger(f"{self.__class__.__name__ }-{self.task_id}")
        self.logger.info('### Configuring task object ###')
        self.dag = dag
        self.dag_id = self.dag.dag_id
        self.repository_url = piece["repository_url"]
        self.repository_version = piece["repository_version"]
        self.piece = piece
        self.piece_input_kwargs = piece_input_kwargs
        self.username_var = username
        self.task_id_main_branch = task_id_main_branch
        self.task_id_else_branch = task_id_else_branch
        self.trigger_rule = trigger_rule
        if "execution_mode" not in self.piece:
            self.execution_mode = "docker"
        else:
            self.execution_mode = self.piece["execution_mode"]

        # Shared storage
        if not workflow_shared_storage:
            workflow_shared_storage = {}
        shared_storage_source_name = StorageSource(workflow_shared_storage.pop("source", "None")).name
        provider_options = workflow_shared_storage.pop("provider_options", {})
        if shared_storage_map[shared_storage_source_name]:
            self.workflow_shared_storage = shared_storage_map[shared_storage_source_name](
                **workflow_shared_storage,
                **provider_options
            )
        else:
            self.workflow_shared_storage = shared_storage_map[shared_storage_source_name]

        # Container resources
        self.container_resources = container_resources

        # Get deploy mode
        self.deploy_mode = os.environ.get('STARFLOW_DEPLOY_MODE', 'prod')

        # Set up task operator
        self._task_operator = self._set_operator()

    def _set_operator(self) -> BaseOperator:
        """
        Set Airflow Operator according to deploy mode and Piece execution mode.
        """

        # References:
        # - https://airflow.apache.org/docs/apache-airflow/1.10.14/_api/airflow/contrib/operators/kubernetes_pod_operator/index.html
        # - https://airflow.apache.org/docs/apache-airflow/stable/templates-ref.html
        # - https://www.astronomer.io/guides/templating/
        # - good example: https://github.com/apache/airflow/blob/main/tests/system/providers/cncf/kubernetes/example_kubernetes.py
        # - commands HAVE to go in a list object: https://stackoverflow.com/a/55149915/11483674

        return starflowKubernetesBranchPodOperator(
            dag_id=self.dag_id,
            task_id=self.task_id,
            piece_name=self.piece.get('name'),
            deploy_mode=self.deploy_mode,
            repository_url=self.repository_url,
            repository_version=self.repository_version,
            workspace_id=self.workspace_id,
            task_id_main_branch=self.task_id_main_branch,
            task_id_else_branch=self.task_id_else_branch,
            piece_input_kwargs=self.piece_input_kwargs,
            workflow_shared_storage=self.workflow_shared_storage,
            container_resources=self.container_resources,
            # ----------------- Kubernetes -----------------
            namespace='airflow',
            image=self.piece.get("source_image"),
            image_pull_policy='IfNotPresent',
            name=f"airflow-worker-pod-{self.task_id.lower().replace('_', '-')}",
            startup_timeout_seconds=600,
            annotations={"sidecar.istio.io/inject": "false"}, # TODO - remove this when istio is working with airflow k8s pod
            # cmds=["/bin/bash"],
            # arguments=["-c", "sleep 120;"],
            cmds=["starflow"],
            arguments=["run-piece-branch-k8s"],
            do_xcom_push=True,
            in_cluster=True,
            username=self.username_var,
            volume_mounts=[
                k8s.V1VolumeMount(
                    name='volume-prod', mount_path='/var/mount_secrets', sub_path=None, read_only=True
                )
            ],
            volumes=[
                k8s.V1Volume(
                    name="volume-prod",
                    secret=k8s.V1SecretVolumeSource(
                        secret_name="airflow-worker-secrets"
                    )
                )
            ],
            trigger_rule=self.trigger_rule
        )
            

    def __call__(self) -> Callable:
        return self._task_operator



from pathlib import Path
import shutil


class LocalFilesClient(object):

    @staticmethod
    def local_import(module_name):
        components = module_name.split('.')
        module = __import__(components[0])
        for comp in components[1:]:
            module = getattr(module, comp)
        return module

    @staticmethod
    def import_module(module_name):
        return LocalFilesClient.local_import(module_name)

    @staticmethod
    def list_files(self, path):
        return [e.name for e in Path(path).iterdir() if e.is_dir() and e.name != '__pycache__']

    @staticmethod
    def save_text_file(path, content):
        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(content)
    
    @staticmethod
    def copy_file(source, target):
        shutil.copyfile(source, target)


COLOR_PALETTE = {
    "success": "rgb(113,176,109)",
    "warning": "rgb(176,169,109)",
    "error": "rgb(176,109,109)"
}

STARFLOW_HELM_PATH = 'starflow/starflow'
STARFLOW_HELM_VERSION = '0.1.6'
STARFLOW_HELM_REPOSITORY = 'https://prochain-star-atlas.github.io/starflow/'


from pydantic import BaseModel, Field


class SystemRequirementsModel(BaseModel):
    cpu: str = Field(pattern=r"^\d+\.*\d*m$")
    memory: str = Field(pattern=r"^\d+\.*\d*Mi$")


class ContainerResourcesModel(BaseModel):
    requests: SystemRequirementsModel = Field(default=SystemRequirementsModel(cpu="100m", memory="128Mi"))
    limits: SystemRequirementsModel = Field(default=SystemRequirementsModel(cpu="100m", memory="128Mi"))
    use_gpu: bool = False


import os
import tomli
import tomli_w
import yaml
import subprocess
import re
import shutil
import requests
import time
from concurrent.futures import ThreadPoolExecutor
import base64
from pathlib import Path
from rich.console import Console
from yaml.resolver import BaseResolver
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from tempfile import NamedTemporaryFile, TemporaryDirectory
from kubernetes import client, config

from starflow.cli.utils.constants import COLOR_PALETTE, STARFLOW_HELM_PATH, STARFLOW_HELM_VERSION, STARFLOW_HELM_REPOSITORY


class AsLiteral(str):
    pass


def represent_literal(dumper, data):
    return dumper.represent_scalar(BaseResolver.DEFAULT_SCALAR_TAG, data, style="|")


yaml.add_representer(AsLiteral, represent_literal)


console = Console()


def create_ssh_pair_key() -> None:
    # Create SSH key pair for GitHub Workflows
    console.print("Generating SSH key pair for GitHub Workflows...")
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=4096
    )

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption()
    )
    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH
    )
    return private_key, public_key


def prepare_platform(
    cluster_name: str,
    workflows_repository: str,
    github_workflows_ssh_private_key: str,
    github_default_pieces_repository_token: str,
    github_workflows_token: str,
    deploy_mode: str,
    local_pieces_repository_path: list,
    local_starflow_path: str,
    local_rest_image: str,
    local_frontend_image: str,
    local_airflow_image: str,
) -> None:
    # Create local configuration file updated with user-provided arguments
    config_file_path = Path(__file__).resolve().parent / "config-starflow-local.toml"
    with open(str(config_file_path), "rb") as f:
        config_dict = tomli.load(f)

    running_path = str(Path().cwd().resolve())
    config_dict["path"]["STARFLOW_LOCAL_RUNNING_PATH"] = running_path
    config_dict["kind"]["STARFLOW_KIND_CLUSTER_NAME"] = cluster_name
    config_dict['kind']['STARFLOW_DEPLOY_MODE'] = deploy_mode

    if deploy_mode == 'local-k8s-dev':
        config_dict['dev']['STARFLOW_AIRFLOW_IMAGE'] = local_airflow_image
        config_dict['dev']['STARFLOW_REST_IMAGE'] = local_rest_image
        config_dict['dev']['STARFLOW_FRONTEND_IMAGE'] = local_frontend_image
        config_dict['dev']['STARFLOW_LOCAL_STARFLOW_PACKAGE'] = local_starflow_path
        for local_pieces_repository in local_pieces_repository_path:
            # Read repo config.toml to get repo name to map it to cluster path
            repo_config_file_path = Path(local_pieces_repository).resolve() / "config.toml"
            with open(str(repo_config_file_path), "rb") as f:
                repo_toml = tomli.load(f)

            repo_name = repo_toml['repository']['REPOSITORY_NAME']
            config_dict['dev'][repo_name] = local_pieces_repository

    config_dict['github']['STARFLOW_GITHUB_WORKFLOWS_REPOSITORY'] = workflows_repository.split("github.com/")[-1].strip('/')

    if not github_workflows_ssh_private_key:
        private_key, public_key = create_ssh_pair_key()
        config_dict["github"]["STARFLOW_GITHUB_WORKFLOWS_SSH_PRIVATE_KEY"] = base64.b64encode(private_key).decode('utf-8')
        config_dict["github"]["STARFLOW_GITHUB_WORKFLOWS_SSH_PUBLIC_KEY"] = public_key.decode("utf-8")
    else:
        config_dict["github"]["STARFLOW_GITHUB_WORKFLOWS_SSH_PRIVATE_KEY"] = github_workflows_ssh_private_key

    config_dict['github']['STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS'] = github_workflows_token
    config_dict['github']['STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN'] = github_default_pieces_repository_token

    with open("config-starflow-local.toml", "wb") as f:
        tomli_w.dump(config_dict, f)

    console.print("")
    console.print(f"starflow is prepared to run at: {running_path}")
    console.print(f"You can check and modify the configuration file at: {running_path}/config-starflow-local.toml")
    console.print("Next, run: `starflow platform create`")
    console.print("")


def create_platform(install_airflow: bool = True, use_gpu: bool = False) -> None:
    # Load configuration values
    with open("config-starflow-local.toml", "rb") as f:
        platform_config = tomli.load(f)

    # Create kind config file and run bash script to create Kind cluster
    kubeadm_config_patches = dict(
        kind="InitConfiguration",
        nodeRegistration=dict(
            kubeletExtraArgs={
                "node-labels": "ingress-ready=true"
            }
        )
    )
    extra_mounts_local_repositories = []

    starflow_dev_private_variables_list = [
        "STARFLOW_LOCAL_STARFLOW_PACKAGE",
        "STARFLOW_REST_IMAGE",
        "STARFLOW_FRONTEND_IMAGE",
        "STARFLOW_AIRFLOW_IMAGE"
    ]
    local_pieces_respositories = {key: value for key, value in platform_config['dev'].items() if key not in starflow_dev_private_variables_list}
    if platform_config['kind']['STARFLOW_DEPLOY_MODE'] == 'local-k8s-dev':
        for repo_name, repo_path in local_pieces_respositories.items():
            extra_mounts_local_repositories.append(
                dict(
                    hostPath=repo_path,
                    containerPath=f"/pieces_repositories/{repo_name}",
                    readOnly=True,
                    propagation='HostToContainer'
                )
            )
        if platform_config['dev'].get('STARFLOW_LOCAL_STARFLOW_PACKAGE'):
            starflow_local_package_absolute_path = Path(platform_config['dev']['STARFLOW_LOCAL_STARFLOW_PACKAGE']).resolve()
            extra_mounts_local_repositories.append(
                dict(
                    hostPath=str(starflow_local_package_absolute_path),
                    containerPath=f"/starflow/starflow_py/src/starflow",
                    readOnly=True,
                    propagation='HostToContainer'
                )
            )

    kubeadm_parsed = AsLiteral(yaml.dump(kubeadm_config_patches))
    use_gpu_dict = {} if not use_gpu else {"gpus": True}
    kind_config = dict(
        kind="Cluster",
        apiVersion="kind.x-k8s.io/v1alpha4",
        nodes=[
            dict(
                role="control-plane",
                kubeadmConfigPatches=[kubeadm_parsed],
                extraPortMappings=[
                    dict(
                        containerPort=80,
                        hostPort=80,
                        listenAddress="0.0.0.0",
                        protocol="TCP"
                    ),
                    dict(
                        containerPort=443,
                        hostPort=443,
                        listenAddress="0.0.0.0",
                        protocol="TCP"
                    )
                ]
            ),
            dict(
                role="worker",
                extraMounts=[
                    dict(
                        hostPath=platform_config["path"]["STARFLOW_LOCAL_RUNNING_PATH"] + "/workflow_shared_storage",
                        containerPath="/cluster_shared_storage",
                        readOnly=False,
                        propagation="Bidirectional"
                    ),
                    *extra_mounts_local_repositories
                ],
                **use_gpu_dict
            ),
        ]
    )
    with open("kind-cluster-config.yaml", "w") as f:
        yaml.dump(kind_config, f)

    cluster_name = platform_config["kind"]["STARFLOW_KIND_CLUSTER_NAME"]

    # Delete previous Kind cluster
    console.print("")
    console.print(f"Removing previous Kind cluster - {cluster_name}...")
    result = subprocess.run(["kind", "delete", "cluster", "--name", cluster_name], capture_output=True, text=True)
    if result.returncode != 0:
        error_message = result.stderr.strip() if result.stderr else result.stdout.strip()
        raise Exception(f"An error occurred while deleting previous Kind cluster - {cluster_name}: {error_message}")
    console.print("")

    # Create new Kind cluster
    console.print(f"Creating new Kind cluster - {cluster_name}...")
    result = subprocess.run(["kind", "create", "cluster", "--name", cluster_name, "--config", "kind-cluster-config.yaml"])
    if result.returncode != 0:
        error_message = result.stderr.strip() if result.stderr else result.stdout.strip()
        raise Exception(f"An error occurred while creating Kind cluster - {cluster_name}: {error_message}")
    console.print("")
    console.print("Kind cluster created successfully!", style=f"bold {COLOR_PALETTE.get('success')}")

    # Install Ingress NGINX controller
    console.print("")
    console.print("Installing NGINX controller...")
    subprocess.run(["kubectl", "apply", "-f", "https://raw.githubusercontent.com/kubernetes/ingress-nginx/main/deploy/static/provider/kind/deploy.yaml"], stdout=subprocess.DEVNULL)
    result = subprocess.run(["kubectl", "wait", "--namespace", "ingress-nginx", "--for", "condition=ready", "pod", "--selector=app.kubernetes.io/component=controller", "--timeout=660s"])
    if result.returncode != 0:
        error_message = result.stderr.strip() if result.stderr else result.stdout.strip()
        raise Exception("An error occurred while installing NGINX controller: {error_message}")
    console.print("NGINX controller installed successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
    console.print("")

    # Load local images to Kind cluster
    local_starflow_airflow_image = platform_config.get('dev', {}).get('STARFLOW_AIRFLOW_IMAGE', None)
    local_starflow_frontend_image = platform_config.get('dev', {}).get('STARFLOW_FRONTEND_IMAGE', None)
    local_starflow_rest_image = platform_config.get('dev', {}).get('STARFLOW_REST_IMAGE', None)

    starflow_airflow_image_tag = 'latest'
    starflow_airflow_image = "ghcr.io/prochain-star-atlas/starflow-airflow-base"
    if local_starflow_airflow_image:
        console.print(f"Loading local starflow Airflow image {local_starflow_airflow_image} to Kind cluster...")
        subprocess.run(["kind", "load", "docker-image", local_starflow_airflow_image , "--name", cluster_name, "--nodes", f"{cluster_name}-worker"])
        starflow_airflow_image = f'docker.io/library/{local_starflow_airflow_image}'
    elif platform_config['kind']["STARFLOW_DEPLOY_MODE"] == 'local-k8s-dev' and not local_starflow_airflow_image:
        starflow_airflow_image_tag = 'latest-dev'

    if local_starflow_frontend_image:
        console.print(f"Loading local frontend image {local_starflow_frontend_image} to Kind cluster...")
        subprocess.run(["kind", "load", "docker-image", local_starflow_frontend_image , "--name", cluster_name, "--nodes", f"{cluster_name}-worker"])
        starflow_frontend_image = f"docker.io/library/{local_starflow_frontend_image}"
    elif platform_config['kind']["STARFLOW_DEPLOY_MODE"] == 'local-k8s-dev':
        starflow_frontend_image = "ghcr.io/prochain-star-atlas/starflow-frontend:k8s-dev"
    else:
        starflow_frontend_image = "ghcr.io/prochain-star-atlas/starflow-frontend:k8s"

    if local_starflow_rest_image:
        console.print(f"Loading local REST image {local_starflow_rest_image} to Kind cluster...")
        subprocess.run(["kind", "load", "docker-image", local_starflow_rest_image , "--name", cluster_name, "--nodes", f"{cluster_name}-worker"])
        starflow_rest_image = f'docker.io/library/{local_starflow_rest_image}'
    elif platform_config['kind']["STARFLOW_DEPLOY_MODE"] == 'local-k8s-dev':
        starflow_rest_image = "ghcr.io/prochain-star-atlas/starflow-rest:latest-dev"
    else:
        starflow_rest_image = "ghcr.io/prochain-star-atlas/starflow-rest:latest"

    # In order to use nvidia gpu in our cluster we need nvidia plugins to be installed.
    # We can use nvidia operator to install nvidia plugins.
    # References:
    #     https://catalog.ngc.nvidia.com/orgs/nvidia/containers/gpu-operator
    #     https://jacobtomlinson.dev/posts/2022/quick-hack-adding-gpu-support-to-kind/
    if use_gpu:
        console.print("Installing NVIDIA GPU Operator...")
        nvidia_gpu_operator_add_repo_command = [
            "helm", "repo", "add", "nvidia", "https://nvidia.github.io/gpu-operator",
        ]
        subprocess.run(nvidia_gpu_operator_add_repo_command)
        helm_update_command = ["helm", "repo", "update"]
        subprocess.run(helm_update_command)

        # We don't need driver as we are using kind and our host machine already has nvidia driver that is why we are disabling it.
        nvidia_plugis_install_command = "helm install --wait --generate-name -n gpu-operator --create-namespace nvidia/gpu-operator --set driver.enabled=false"
        subprocess.run(nvidia_plugis_install_command, shell=True)


    # Override values for starflow Helm chart
    db_enabled = platform_config['starflow_db'].get("STARFLOW_CREATE_DATABASE", True)
    token_pieces = platform_config["github"]["STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN"]
    token_workflows = platform_config["github"]["STARFLOW_GITHUB_ACCESS_TOKEN_WORKFLOWS"]
    starflow_values_override_config = {
        "github_access_token_pieces": token_pieces,
        "github_access_token_workflows": token_workflows,
        "frontend": {
            "enabled": True,
            "image": starflow_frontend_image,
            "apiEnv": "dev" if platform_config['kind']["STARFLOW_DEPLOY_MODE"] in ['local-k8s-dev', 'local-k8s'] else 'prod',
            "deployMode": platform_config['kind']["STARFLOW_DEPLOY_MODE"],
        },
        "rest": {
            "enabled": True,
            "image": starflow_rest_image,
            "workflowsRepository": platform_config['github']['STARFLOW_GITHUB_WORKFLOWS_REPOSITORY'],
            "createDefaultUser": platform_config['starflow_db'].get('STARFLOW_CREATE_DEFAULT_USER', True)
        },
        "database": {
            "enabled": db_enabled,
            "image": "postgres:13",
            "name": "postgres",
            "user": "postgres",
            "password": "postgres",
            "port": "5432",
        }
    }

    # Only add database values if database is enabled
    # If not enabled will use always the default values
    if not db_enabled:
        starflow_values_override_config['database'] = {
            **starflow_values_override_config['database'],
            "host": platform_config['starflow_db']["STARFLOW_DB_HOST"],
            "name":  platform_config['starflow_db']["STARFLOW_DB_NAME"],
            "user": platform_config['starflow_db']["STARFLOW_DB_USER"],
            "password": platform_config['starflow_db']["STARFLOW_DB_PASSWORD"],
            "port": str(platform_config['starflow_db'].get("STARFLOW_DB_PORT", 5432))
        }

    # Override values for Airflow Helm chart
    airflow_ssh_config = dict(
        gitSshKey=f"{platform_config['github']['STARFLOW_GITHUB_WORKFLOWS_SSH_PRIVATE_KEY']}",
    )
    airflow_ssh_config_parsed = AsLiteral(yaml.dump(airflow_ssh_config))

    workers_extra_volumes = []
    workers_extra_volumes_mounts = []
    workers = {}
    scheduler = {}
    if platform_config['kind']["STARFLOW_DEPLOY_MODE"] == 'local-k8s-dev' and platform_config['dev'].get('STARFLOW_LOCAL_STARFLOW_PACKAGE'):
        workers_extra_volumes = [
            {
                "name": "starflow-dev-extra",
                "persistentVolumeClaim": {
                    "claimName": "starflow-dev-volume-claim"
                }
            }
        ]
        workers_extra_volumes_mounts = [
            {
                "name": "starflow-dev-extra",
                "mountPath": "/opt/airflow/starflow/starflow_py/src/starflow"
            }
        ]
        workers = {
            "workers": {
                "extraVolumes": workers_extra_volumes,
                "extraVolumeMounts": workers_extra_volumes_mounts,
            }
        }
        scheduler = {
            "scheduler": {
                "extraVolumes": workers_extra_volumes,
                "extraVolumeMounts": workers_extra_volumes_mounts,
            }
        }

    airflow_values_override_config = {
        "env": [
            {
                "name": "STARFLOW_DEPLOY_MODE",
                "value": platform_config['kind']["STARFLOW_DEPLOY_MODE"]
            },
        ],
        "images": {
            "useDefaultImageForMigration": False,
            "airflow": {
                "repository": starflow_airflow_image,
                "tag": starflow_airflow_image_tag,
                "pullPolicy": "IfNotPresent"
            }
        },
        "extraSecrets": {
            "airflow-ssh-secret": {
                "data": airflow_ssh_config_parsed
            }
        },
        "config": {
            "api": {
                "auth_backends": "airflow.api.auth.backend.basic_auth"
            },
        },
        "dags": {
            "gitSync": {
                "enabled": True,
                "wait": 60,
                "repo": f"ssh://git@github.com/{platform_config['github']['STARFLOW_GITHUB_WORKFLOWS_REPOSITORY']}.git",
                "branch": "main",
                "subPath": "workflows",
                "sshKeySecret": "airflow-ssh-secret"
            },
        },
        "migrateDatabaseJob": {
            "jobAnnotations": {
                "sidecar.istio.io/inject": "false"
            },
            "annotations": {
                "sidecar.istio.io/inject": "false"
            },
        },
        "createUserJob": {
            "jobAnnotations": {
                "sidecar.istio.io/inject": "false"
            },
            "annotations": {
                "sidecar.istio.io/inject": "false"
            },
        },
        **workers,
        **scheduler,
    }

    # Update Helm repositories
    subprocess.run(["helm", "repo", "add", "starflow", STARFLOW_HELM_REPOSITORY])
    subprocess.run(["helm", "repo", "add", "apache-airflow", "https://airflow.apache.org/"])  # ref: https://github.com/helm/helm/issues/8036
    subprocess.run(["helm", "repo", "update"])

    # Install Airflow Helm Chart
    if install_airflow:
        console.print('Installing Apache Airflow...')
        # Create temporary file with airflow values
        with NamedTemporaryFile(suffix='.yaml', mode="w") as fp:
            yaml.dump(airflow_values_override_config, fp)
            commands = [
                "helm", "install",
                "-f", str(fp.name),
                "airflow",
                "apache-airflow/airflow",
                "--version", " 1.11.0",
            ]
            subprocess.run(commands)

    # Install starflow Helm Chart
    local_starflow_path = platform_config.get('dev', {}).get('STARFLOW_LOCAL_STARFLOW_PACKAGE')
    if platform_config.get('kind', {}).get('STARFLOW_DEPLOY_MODE') == 'local-k8s-dev' and local_starflow_path:
        console.print('Installing starflow using local helm...')
        helm_starflow_path = Path(local_starflow_path).parent.parent / "helm/starflow"
        with NamedTemporaryFile(suffix='.yaml', mode="w") as fp:
            yaml.dump(starflow_values_override_config, fp)
            commands = [
                "helm", "install",
                "-f", str(fp.name),
                "starflow",
                helm_starflow_path
            ]
            subprocess.run(commands)
    else:
        console.print('Installing starflow using remote helm...')
        with TemporaryDirectory() as tmp_dir:
            console.print("Downloading starflow Helm chart...")
            subprocess.run([
                "helm",
                "pull",
                STARFLOW_HELM_PATH,
                "--untar",
                "-d",
                tmp_dir
            ])
            with NamedTemporaryFile(suffix='.yaml', mode="w") as fp:
                yaml.dump(starflow_values_override_config, fp)
                console.print('Installing starflow...')
                commands = [
                    "helm", "install",
                    "-f", str(fp.name),
                    "starflow",
                    f"{tmp_dir}/starflow",
                ]
                subprocess.run(commands)

    # For each path create a pv and pvc
    if platform_config['kind']['STARFLOW_DEPLOY_MODE'] == 'local-k8s-dev':
        config.load_kube_config()
        k8s_client = client.CoreV1Api()
        v1 = client.RbacAuthorizationV1Api()

        # Create service account role binding with admin access for airflow worker
        role_binding_name_worker = "full-access-user-clusterrolebinding-worker"
        sa_name = "airflow-worker"
        cluster_role_binding_worker = client.V1ClusterRoleBinding(
            metadata=client.V1ObjectMeta(name=role_binding_name_worker),
            subjects=[
                client.V1Subject(
                    kind="ServiceAccount",
                    name=sa_name,
                    namespace="default"
                )
            ],
            role_ref=client.V1RoleRef(
                kind="ClusterRole",
                name="cluster-admin",
                api_group="rbac.authorization.k8s.io"
            )
        )
        console.print('Creating RBAC Worker Authorization for local dev')
        v1.create_cluster_role_binding(cluster_role_binding_worker)

        role_binding_name_scheduler = "full-access-user-clusterrolebinding-scheduler"
        sa_name = "airflow-scheduler"
        cluster_role_binding_scheduler = client.V1ClusterRoleBinding(
            metadata=client.V1ObjectMeta(name=role_binding_name_scheduler),
            subjects=[
                client.V1Subject(
                    kind="ServiceAccount",
                    name=sa_name,
                    namespace="default"
                )
            ],
            role_ref=client.V1RoleRef(
                kind="ClusterRole",
                name="cluster-admin",
                api_group="rbac.authorization.k8s.io"
            )
        )
        console.print('Creating RBAC Scheduler Authorization for local dev')
        v1.create_cluster_role_binding(cluster_role_binding_scheduler)

        for project_name in local_pieces_respositories.keys():
            console.log(f"Creating PV and PVC for {project_name}...")
            # Check if pv already exists
            persistent_volume_name = 'pv-{}'.format(str(project_name.lower().replace('_', '-')))
            persistent_volume_claim_name = 'pvc-{}'.format(str(project_name.lower().replace('_', '-')))
            pvc_exists = False
            try:
                k8s_client.read_namespaced_persistent_volume_claim(name=persistent_volume_claim_name, namespace='default')
                pvc_exists = True
            except client.rest.ApiException as e:
                if e.status != 404:
                    raise e

            if not pvc_exists:
                pvc = client.V1PersistentVolumeClaim(
                    metadata=client.V1ObjectMeta(name=persistent_volume_claim_name),
                    spec=client.V1PersistentVolumeClaimSpec(
                        access_modes=["ReadOnlyMany"],
                        volume_name=persistent_volume_name,
                        resources=client.V1ResourceRequirements(
                            requests={"storage": "300Mi"}
                        ),
                        storage_class_name="standard"
                    )
                )
                k8s_client.create_namespaced_persistent_volume_claim(namespace="default", body=pvc)

            pv_exists = False
            try:
                k8s_client.read_persistent_volume(name=persistent_volume_name)
                pv_exists = True
            except client.rest.ApiException as e:
                if e.status != 404:
                    raise e

            if not pv_exists:
                pv = client.V1PersistentVolume(
                    metadata=client.V1ObjectMeta(name=persistent_volume_name),
                    spec=client.V1PersistentVolumeSpec(
                        access_modes=["ReadWriteOnce"],
                        capacity={"storage": "1Gi"},
                        persistent_volume_reclaim_policy="Retain",
                        storage_class_name="standard",
                        host_path=client.V1HostPathVolumeSource(path=f"/pieces_repositories/{project_name}"),
                        claim_ref=client.V1ObjectReference(
                            namespace="default",
                            name=persistent_volume_claim_name,
                            kind="PersistentVolumeClaim"
                        ),
                        node_affinity=client.V1VolumeNodeAffinity(
                            required=client.V1NodeSelector(
                                node_selector_terms=[
                                    client.V1NodeSelectorTerm(
                                        match_expressions=[
                                            client.V1NodeSelectorRequirement(
                                                key="kubernetes.io/hostname",
                                                operator="In",
                                                values=["starflow-cluster-worker"]
                                            )
                                        ]
                                    )
                                ]
                            )
                        )
                    )
                )
                k8s_client.create_persistent_volume(body=pv)

        if platform_config['dev'].get('STARFLOW_LOCAL_STARFLOW_PACKAGE'):
            console.print("Creating PV's and PVC's for Local starflow Package...")
            # Create pv and pvc for local dev starflow
            pvc = client.V1PersistentVolumeClaim(
                metadata=client.V1ObjectMeta(name="starflow-dev-volume-claim"),
                spec=client.V1PersistentVolumeClaimSpec(
                    access_modes=["ReadWriteMany"],
                    volume_name="starflow-dev-volume",
                    resources=client.V1ResourceRequirements(
                        requests={"storage": "300Mi"}
                    ),
                    storage_class_name="standard"
                )
            )
            k8s_client.create_namespaced_persistent_volume_claim(namespace="default", body=pvc)
            pv = client.V1PersistentVolume(
                metadata=client.V1ObjectMeta(name="starflow-dev-volume"),
                spec=client.V1PersistentVolumeSpec(
                    storage_class_name="standard",
                    access_modes=["ReadWriteMany"],
                    capacity={"storage": "2Gi"},
                    host_path=client.V1HostPathVolumeSource(path="/starflow/starflow_py/src/starflow"),
                    claim_ref=client.V1ObjectReference(
                        namespace="default",
                        name="starflow-dev-volume-claim",
                    ),
                    node_affinity=client.V1VolumeNodeAffinity(
                        required=client.V1NodeSelector(
                            node_selector_terms=[
                                client.V1NodeSelectorTerm(
                                    match_expressions=[
                                        client.V1NodeSelectorRequirement(
                                            key="kubernetes.io/hostname",
                                            operator="In",
                                            values=["starflow-cluster-worker"]
                                        )
                                    ]
                                )
                            ]
                        )
                    )
                )
            )
            k8s_client.create_persistent_volume(body=pv)

    console.print("")
    console.print("K8s resources created successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
    console.print("You can now access the starflow frontend at: http://localhost/")
    console.print("starflow's REST API: http://localhost/api/")
    console.print("starflow's REST API Swagger: http://localhost/api/docs")
    console.print("")


def destroy_platform() -> None:
    # Delete Kind cluster
    with open("config-starflow-local.toml", "rb") as f:
        platform_config = tomli.load(f)
    cluster_name = platform_config["kind"]["STARFLOW_KIND_CLUSTER_NAME"]
    console.print(f"Removing Kind cluster - {cluster_name}...")
    result = subprocess.run(["kind", "delete", "cluster", "--name", cluster_name], capture_output=True, text=True)
    if result.returncode != 0:
        error_message = result.stderr.strip() if result.stderr else result.stdout.strip()
        raise Exception(f"An error occurred while deleting Kind cluster - {cluster_name}: {error_message}")
    console.print("")
    console.print("Kind cluster removed successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
    console.print("")


def run_platform_compose(
    github_token: str,
    use_config_file: bool = False,
    dev: bool = False,
    debug: bool = False
) -> None:
    console.print("Starting starflow Platform using Docker Compose.")
    console.print("Please wait, this might take a few minutes...")
    # Database default settings
    create_database = True
    os.environ['STARFLOW_CREATE_DEFAULT_USER'] = 'true'
    os.environ['STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN'] = github_token
    if use_config_file:
        console.print("Using config file...")
        with open("config-starflow-local.toml", "rb") as f:
            platform_config = tomli.load(f)
        create_database = platform_config['starflow_db'].get('STARFLOW_CREATE_DATABASE', True)
        os.environ['STARFLOW_CREATE_DEFAULT_USER'] = str(platform_config['starflow_db'].get('STARFLOW_CREATE_DEFAULT_USER', 'true')).lower()

        if not create_database:
            os.environ['STARFLOW_DB_HOST'] = platform_config['starflow_db'].get("STARFLOW_DB_HOST", 'postgres')
            os.environ['STARFLOW_DB_PORT'] = platform_config['starflow_db'].get("STARFLOW_DB_PORT", 5432)
            os.environ['STARFLOW_DB_USER'] = platform_config['starflow_db'].get("STARFLOW_DB_USER", 'postgres')
            os.environ['STARFLOW_DB_PASSWORD'] = platform_config['starflow_db'].get("STARFLOW_DB_PASSWORD", 'postgres')
            os.environ['STARFLOW_DB_NAME'] = platform_config['starflow_db'].get("STARFLOW_DB_NAME", 'postgres')
            os.environ['NETWORK_MODE'] = 'bridge'

        # If running database in an external local container, set network mode to host
        if platform_config['starflow_db'].get('STARFLOW_DB_HOST') in ['localhost', '0.0.0.0', '127.0.0.1']:
            os.environ['NETWORK_MODE'] = 'host'

    # Create local directories
    local_path = Path(".").resolve()
    starflow_dir = local_path / "starflow_data"
    starflow_dir.mkdir(parents=True, exist_ok=True)
    starflow_dir.chmod(0o777)

    airflow_base = local_path / 'airflow'
    airflow_logs_dir = airflow_base / "logs"
    airflow_logs_dir.mkdir(parents=True, exist_ok=True)
    airflow_dags_dir = airflow_base / "dags"
    airflow_dags_dir.mkdir(parents=True, exist_ok=True)
    airflow_plugins_dir = airflow_base / "plugins"
    airflow_plugins_dir.mkdir(parents=True, exist_ok=True)
    airflow_base.chmod(0o777)

    # Copy docker-compose.yaml file from package to local path
    if create_database:
        docker_compose_path = Path(__file__).resolve().parent / "docker-compose.yaml"
    else:
        docker_compose_path = Path(__file__).resolve().parent / "docker-compose-without-database.yaml"
    shutil.copy(str(docker_compose_path), "./docker-compose.yaml")

    # Environment variables
    environment = os.environ.copy()
    environment['STARFLOW_COMPOSE_DEV'] = ''
    if dev:
        environment['STARFLOW_COMPOSE_DEV'] = '-dev'

    # Run docker compose pull
    console.print("\nPulling Docker images...")
    pull_cmd = [
        "docker",
        "compose",
        "pull"
    ]
    pull_process = subprocess.Popen(pull_cmd, env=environment)
    pull_process.wait()
    if pull_process.returncode == 0:
        console.print(" \u2713 Docker images pulled successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
    else:
        console.print("Docker images pull failed.", style=f"bold {COLOR_PALETTE.get('error')}")

    # Run docker compose up
    console.print("\nStarting services...")
    cmd = [
        "docker",
        "compose",
        "up"
    ]

    if debug:
        subprocess.Popen(cmd, env=environment)
    else:
        airflow_redis_ready = False
        airflow_database_ready = False
        airflow_init_ready = False
        airflow_triggerer_ready = False
        airflow_worker_ready = False
        airflow_webserver_ready = False
        airflow_scheduler_ready = False
        starflow_database_ready = False

        process = subprocess.Popen(cmd, env=environment, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

        # Read and filter the output
        def customize_message(
            line: str,
            airflow_redis_ready: bool = False,
            airflow_database_ready: bool = False,
            airflow_init_ready: bool = False,
            airflow_triggerer_ready: bool = False,
            airflow_worker_ready: bool = False,
            airflow_webserver_ready: bool = False,
            airflow_scheduler_ready: bool = False,
            starflow_database_ready: bool = False,
        ):
            line = line.lower()
            line = re.sub(r'\s+', ' ', line)
            if not airflow_redis_ready and "airflow-redis" in line and "ready to accept connections tcp" in line:
                console.print(" \u2713 Airflow Redis service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_redis_ready = True
            if not airflow_database_ready and "airflow-postgres" in line and ("ready" in line or "skipping" in line):
                console.print(" \u2713 Airflow database service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_database_ready = True
            if not airflow_init_ready and "airflow-init" in line and "exited with code 0" in line:
                console.print(" \u2713 Airflow pre-initialization service completed successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_init_ready = True
            if not airflow_triggerer_ready and "airflow-triggerer" in line and "starting" in line:
                console.print(" \u2713 Airflow triggerer service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_triggerer_ready = True
            if not airflow_worker_ready and "airflow-starflow-worker" in line and "execute_command" in line:
                console.print(" \u2713 Airflow worker service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_worker_ready = True
            if not airflow_webserver_ready and "airflow-webserver" in line and "health" in line and "200" in line:
                console.print(" \u2713 Airflow webserver service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_webserver_ready = True
            if not airflow_scheduler_ready and "airflow-starflow-scheduler" in line and "launched" in line:
                console.print(" \u2713 Airflow scheduler service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                airflow_scheduler_ready = True
            if not starflow_database_ready and "starflow-postgres" in line and ("ready" in line or "skipping" in line):
                console.print(" \u2713 starflow database service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                starflow_database_ready = True
            return airflow_redis_ready, airflow_database_ready, airflow_init_ready, airflow_triggerer_ready, airflow_worker_ready, airflow_webserver_ready, airflow_scheduler_ready, starflow_database_ready

        def check_starflow_processes():
            while True:
                frontend_response = requests.get("http://localhost:3000").status_code
                rest_response = requests.get("http://localhost:8000/health-check").status_code
                if frontend_response == 200 and rest_response == 200:
                    console.print(" \u2713 starflow REST service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                    console.print(" \u2713 starflow frontend service started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                    break
                time.sleep(5)

        for line in process.stdout:
            airflow_redis_ready, airflow_database_ready, airflow_init_ready, airflow_triggerer_ready, airflow_worker_ready, airflow_webserver_ready, airflow_scheduler_ready, starflow_database_ready = customize_message(
                line, airflow_redis_ready, airflow_database_ready, airflow_init_ready, airflow_triggerer_ready,
                airflow_worker_ready, airflow_webserver_ready, airflow_scheduler_ready, starflow_database_ready)
            if all([
                airflow_redis_ready,
                airflow_database_ready,
                airflow_init_ready,
                airflow_triggerer_ready,
                airflow_worker_ready,
                airflow_webserver_ready,
                airflow_scheduler_ready,
                starflow_database_ready,
            ]):
                check_starflow_processes()
                console.print("\n \u2713 All services for starflow Platform started successfully!", style=f"bold {COLOR_PALETTE.get('success')}")
                console.print("")
                console.print("You can now access them at")
                console.print("starflow UI: http://localhost:3000")
                console.print("starflow REST API: http://localhost:8000")
                console.print("starflow REST API Docs: http://localhost:8000/docs")
                console.print("Airflow webserver: http://localhost:8080")
                console.print("")
                console.print("To stop the platform, run:")
                console.print("    $ starflow platform stop-compose")
                console.print("")
                break


def stop_platform_compose() -> None:
    # If "docker-compose.yaml" file is present in current working path, try run "docker compose down"
    docker_compose_path = Path.cwd().resolve() / "docker-compose.yaml"
    if docker_compose_path.exists():
        # Setting this environment variable to empty string just to print cleaner messages to terminal
        environment = os.environ.copy()
        environment['STARFLOW_COMPOSE_DEV'] = ''
        environment['STARFLOW_DEFAULT_PIECES_REPOSITORY_TOKEN'] = ''
        environment['AIRFLOW_UID'] = ''
        cmd = [
            "docker",
            "compose",
            "down"
        ]
        completed_process = subprocess.run(cmd, env=environment)
        # if completed_process.returncode == 0:
        #     print("starflow Platform stopped successfully. All containers were removed.")

    # Stop and remove containers by name
    def stop_and_remove_container(container_name):
        print(f"Stopping {container_name}...")
        process = subprocess.Popen(f"docker stop {container_name}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            # print(f"Command failed with error: {stderr.decode()}")
            pass
        else:
            print(stdout.decode())

        print(f"Removing {container_name}...")
        process = subprocess.Popen(f"docker rm {container_name}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            # print(f"Command failed with error: {stderr.decode()}")
            pass
        else:
            print(stdout.decode())

    try:
        container_names = [
            "starflow-frontend",
            "starflow-rest",
            "starflow-postgres",
            "starflow-docker-proxy",
            "airflow-starflow-scheduler",
            "airflow-starflow-worker",
            "airflow-webserver",
            "airflow-triggerer",
            "airflow-redis",
            "airflow-postgres",
            "airflow-flower",
            "airflow-cli",
            "airflow-init",
        ]
        with ThreadPoolExecutor() as executor:
            executor.map(stop_and_remove_container, container_names)
        console.print("\n \u2713 starflow Platform stopped successfully. All containers were removed.\n", style=f"bold {COLOR_PALETTE.get('success')}")
    except Exception as e:
        print(f"An error occurred: {e}")



from enum import Enum


class DisplayResultFileType(str, Enum):
    png = "png"
    jpeg = "jpeg"
    bmp = "bmp"
    gif = "gif"
    tiff = "tiff"
    svg = "svg"
    txt = "txt"
    json = "json"
    md = "md"
    pdf = "pdf"
    html = "html"
    plotly_json = "plotly_json"


import yaml
import json
import os
from pathlib import Path


# TODO - UPDATE THIS


class MyDumper(yaml.SafeDumper):
    def write_line_break(self, data=None):
        """
        Insert blank lines between top-level objects
        From here: https://github.com/yaml/pyyaml/issues/127#issuecomment-525800484
        inspired by https://stackoverflow.com/a/44284819/3786245
        """
        super().write_line_break(data)
        if len(self.indents) == 1:
            super().write_line_break()
        elif len(self.indents) == 2:
            super().write_line_break()

    def ignore_aliases(self, data):
        return True
    
    def increase_indent(self, flow=False, indentless=False):
        """
        Proper indentation of lists
        From here: https://stackoverflow.com/a/39681672/11483674
        """
        return super(MyDumper, self).increase_indent(flow, False)

    @staticmethod
    def str_presenter(dumper, data):
        """
        Configures yaml for dumping multiline strings
        Ref: https://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data
        """
        if len(data.splitlines()) > 1:  # check for multiline string
            return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
        return dumper.represent_scalar('tag:yaml.org,2002:str', data)


yaml.add_representer(str, MyDumper.str_presenter)
yaml.representer.SafeRepresenter.add_representer(str, MyDumper.str_presenter)


def create_dependencies_services(dependencies_map: dict, run_scope: str):
    default_image = "atlassistant/flowui-airflow-base-worker"

    # This needs to run only after ENV vars have been set
    from starflow.scripts.docker_compose_constants import (
        x_airflow_common,
        airflow_common_env,
        airflow_common_depends_on,
        x_airflow_common_dev
    )

    command_def = {
        "deploy-local": 'bash -c "pip install flowui && airflow celery worker --autoscale 1 --queues {dependency_name}"',
        "deploy-local-dev": 'bash -c "pip install -e /opt/flowui && airflow celery worker --autoscale 1 --queues {dependency_name}"'
    }

    x_airflow_common_def = {
        "deploy-local": x_airflow_common,
        "deploy-local-dev": x_airflow_common_dev,
    }

    services = dict()
    for dependency_name, dependency_info in dependencies_map.items():
        environment_secrets = {i:f"${{{i}:-}}" for i in dependency_info.get("secrets", list())}
        services[f'airflow-{dependency_name}'] = {
            **x_airflow_common_def[run_scope],
            "image": dependency_info.get("source_image", default_image),
            "container_name": dependency_name,
            "command": command_def.get(run_scope).format(dependency_name=dependency_name), #f'bash -c "pip install -e ./flowui/ && airflow celery worker --autoscale 1 --queues {dependency_name}"',
            "healthcheck": {
                "test": [
                    "CMD-SHELL",
                    'celery --app airflow.executors.celery_executor.app inspect ping -d "celery@$${HOSTNAME}"'
                ],
                "interval": "10s",
                "timeout": "10s",
                "retries": 5
            },
            "environment": {
                **airflow_common_env,
                **environment_secrets,
                "DUMB_INIT_SETSID": "0"
            },
            "restart": "always",
            "depends_on": {
                **airflow_common_depends_on,
                "airflow-init": {
                    "condition": "service_completed_successfully"
                }
            }
        }

    return services


# TODO change dependencies_map_path to be based on the mnt directory defined in config.ini
def create_docker_compose_file(
    dependencies_map: dict = None,
    run_scope: str = "deploy-local",
    output_file_path: str = "docker-compose.yaml",
    include_airflow_cli: bool = False,
    include_airflow_triggerer: bool = False,
    include_flower: bool = False,
    include_flowui_backend: bool = True,
    include_flowui_frontend: bool = False
):
    # This needs to run only after ENV vars have been set
    from starflow.scripts.docker_compose_constants import (
        airflow_postgres,
        airflow_redis,
        airflow_scheduler,
        airflow_triggerer,
        airflow_webserver,
        airflow_webserver_dev,
        airflow_cli,
        airflow_init,
        flower,
        flowui_backend,
        flowui_postgres,
        flowui_frontend,
        flowui_backend_dev,
        flowui_frontend_dev,
        airflow_scheduler_dev
    )

    dependencies_services = create_dependencies_services(
        dependencies_map=dependencies_map,
        run_scope=run_scope,
    )

    # Services
    scheduler_def = {
        "deploy-local": airflow_scheduler,
        "deploy-local-dev": airflow_scheduler_dev
    }

    webserver_def = {
        "deploy-local": airflow_webserver,
        "deploy-local-dev": airflow_webserver_dev
    }

    services = {
        "postgres": airflow_postgres,
        "redis": airflow_redis,
        "airflow-webserver": webserver_def[run_scope],
        "airflow-scheduler": scheduler_def[run_scope],
        "airflow-init": airflow_init,
        **dependencies_services
    }

    # Volumes
    volumes = {"postgres-airflow-volume": None}

    # Networks
    networks = {}

    # Conditional items
    if include_airflow_cli:
        services.update({"airflow-cli": airflow_cli})
    if include_airflow_triggerer:
        services.update({"airflow-triggerer": airflow_triggerer})
    if include_flower:
        services.update({"flower": flower})

    if include_flowui_backend:
        backend_def = {
            "deploy-local": flowui_backend,
            "deploy-local-dev": flowui_backend_dev
        }
        services.update({"flowui-backend": backend_def[run_scope]})
        services.update({"flowui-postgres": flowui_postgres})
        volumes.update({"flowui-postgres-volume": None})
        networks.update({"flowui-postgres-network": {"driver": "bridge"}})

    if include_flowui_frontend:
        frontend_def = {
            "deploy-local": flowui_frontend,
            "deploy-local-dev": flowui_frontend_dev
        }
        services.update({"flowui-frontend": frontend_def[run_scope]})

    file_data = {
        "version": "3",
        "services": services,
        "volumes": volumes,
        "networks": networks
    }

    with open(output_file_path, 'w') as f:
        yaml.dump(
            file_data, 
            f, 
            Dumper=MyDumper,
            indent=4,
            default_style=None, 
            default_flow_style=False, 
            sort_keys=False
        )


if __name__ == '__main__':
    create_docker_compose_file()


init_script = """echo 'Initializing Airflow'
chown -R "${AIRFLOW_UID}:0" /opt/mnt/fs/airflow
function ver() {
  printf "%04d%04d%04d%04d" $${1//./ }
}
airflow_version=$$(gosu airflow airflow version)
airflow_version_comparable=$$(ver $${airflow_version})
min_airflow_version=2.2.0
min_airflow_version_comparable=$$(ver $${min_airflow_version})
if (( airflow_version_comparable < min_airflow_version_comparable )); then
  echo
  echo -e "[Error!!! Too old Airflow version $${airflow_version}!"
  echo "The minimum Airflow version supported: $${min_airflow_version}. Only use this or higher!"
  echo
  exit 1
fi
if [[ -z "${AIRFLOW_UID}" ]]; then
    echo
    echo -e "[WARNING!!!: AIRFLOW_UID not set!"
    echo "If you are on Linux, you SHOULD follow the instructions below to set "
    echo "AIRFLOW_UID environment variable, otherwise files will be owned by root."
    echo "For other operating systems you can get rid of the warning with manually created .env file:"
    echo "    See: https://airflow.apache.org/docs/apache-airflow/stable/start/docker.html#setting-the-right-airflow-user"
    echo
fi
one_meg=1048576
mem_available=$$(($$(getconf _PHYS_PAGES) * $$(getconf PAGE_SIZE) / one_meg))
cpus_available=$$(grep -cE 'cpu[0-9]+' /proc/stat)
disk_available=$$(df / | tail -1 | awk '{print $$4}')
warning_resources="false"
if (( mem_available < 4000 )) ; then
  echo
  echo -e "[WARNING!!!: Not enough memory available for Docker."
  echo "At least 4GB of memory required. You have $$(numfmt --to iec $$((mem_available * one_meg)))"
  echo
  warning_resources="true"
fi
if (( cpus_available < 2 )); then
  echo
  echo -e "[WARNING!!!: Not enough CPUS available for Docker."
  echo "At least 2 CPUs recommended. You have $${cpus_available}"
  echo
  warning_resources="true"
fi
if (( disk_available < one_meg * 10 )); then
  echo
  echo -e "[WARNING!!!: Not enough Disk space available for Docker."
  echo "At least 10 GBs recommended. You have $$(numfmt --to iec $$((disk_available * 1024 )))"
  echo
  warning_resources="true"
fi
if [[ $${warning_resources} == "true" ]]; then
  echo
  echo -e "[WARNING!!!: You have not enough resources to run Airflow (see above)!"
  echo "Please follow the instructions to increase amount of resources available:"
  echo "   https://airflow.apache.org/docs/apache-airflow/stable/start/docker.html#before-you-begin"
  echo
fi
mkdir -p /opt/mnt/fs/airflow/logs /opt/mnt/fs/airflow/dags /opt/mnt/fs/airflow/plugins
chown -R "${AIRFLOW_UID}:0" /opt/mnt/fs/airflow/{logs,dags,plugins}
exec /entrypoint airflow version
"""


metadata_default = {
    "name": "DefaultPiece",
    "dependency": {
        "dockerfile": None,
        "requirements_file": None
    },
    "tags": [],
    "style": {
        "node_label": "Piece",
        "node_type": "default",
        "node_style": {
            "backgroundColor": "#ebebeb"
        },
        "useIcon": True,
        "icon_class_name": "fa-solid fa-circle",
        "iconStyle": {
            "cursor": "pointer"
        }
    },
}


import requests
from requests.exceptions import ConnectionError, Timeout
import time
import docker
from starflow.logger import get_configured_logger
from pathlib import Path


class TestingHttpClient:

    docker_client = docker.from_env()
    STARFLOW_TESTING_CONTAINER_NAME = 'starflow_testing_http_server'
    STARFLOW_INTERNAL_REPOSITORY_FOLDER_PATH = "/home/starflow/pieces_repository/"
    BASE_HTTP_SERVER_HOST_URL = "http://0.0.0.0:8080"
    logger = get_configured_logger("TestingHttpClient")

    @classmethod
    def wait_health_check(cls):
        max_retries = 30
        retry_delay_in_seconds = 1
        url = f'{cls.BASE_HTTP_SERVER_HOST_URL}/health-check'
        for _ in range(max_retries):
            try:
                cls.logger.info(f"Sending health check request to {url}")
                response = requests.get(url)
                response.raise_for_status()
                return response
            except (ConnectionError, Timeout) as e:
                cls.logger.info(f"Health Check Failed with error: {e}.\nRetrying in {retry_delay_in_seconds} seconds...")
                time.sleep(retry_delay_in_seconds)
        raise Exception("Max retries exceeded. Unable to make the request.")

    @classmethod
    def start_http_server(cls, image: str):
        try:
            container = cls.docker_client.containers.run(
                image=image,
                command=["bash", "-c", "python -c 'from starflow.testing import http_server; http_server.run_server()'"],
                ports={'8080/tcp': ('0.0.0.0', 8080)},
                name=cls.STARFLOW_TESTING_CONTAINER_NAME,
                detach=True,
            )
            container_state = container.attrs.get('State').get('Running')
            container_status = container.attrs.get('State').get('Status', None)
            while not container_state and container_status != 'running':
                container = cls.docker_client.containers.get(container.id)
                container_state = container.attrs.get('State').get('Running')
                container_status = container.attrs.get('State').get('Status')
                cls.logger.info('Waiting for container to start...')
                time.sleep(0.2)
            cls.wait_health_check()
            return container
        except Exception as e:
            container = cls.docker_client.containers.get(cls.STARFLOW_TESTING_CONTAINER_NAME)
            if container:
                container.stop()
                container.remove()
            raise e

    @classmethod
    def send_dry_run_request(
        cls,
        piece_name: str,
        input_data: dict,
        secrets_data: dict
    ):
        url = f"{cls.BASE_HTTP_SERVER_HOST_URL}/test"
        body_data = dict(
            piece_name=piece_name,
            input_data=input_data,
            secrets_data=secrets_data,
            repository_folder_path=cls.STARFLOW_INTERNAL_REPOSITORY_FOLDER_PATH
        )
        cls.logger.info(f'Sending dry run request - \n{body_data}')

        response = requests.post(url, json=body_data)
        return response
